# Copyright (C) 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import asyncio

import logging
from .dependency_injection import AsyncInjectable, inject_autokwargs, InjectionKey, inject, AsyncInjector
from .network import NetworkLink

__all__ = []

logger = logging.getLogger('carthage.network')


class DnsZone(AsyncInjectable):

    def __init__(self, name=None, **kwargs):
        if name:
            self.name = name
        super().__init__(**kwargs)
        if not getattr(self, 'name', None):
            raise TypeError('Name must be specified or set in a subclass')

    def contains(self, name):
        '''
        Returns `bool` representing whether or not zone should contain name
        '''
        # we trim the trailing dot that is returned from the API
        # so we just trim the dot on the fqdn we are passed if it has one
        if name.endswith('.'):
            name = name[:-1]
        return name.endswith(self.name)

    async def update_records(self, *args, ttl=300):
        '''
        Updates in a DNS Zone record(s)
        Arguments::
            *args : must be sequences representing records
            record (sequence) : (Name, type, Value) must be specified
                Value may be list or str

        Typical usage::
            zone.update_records(
                    ('foo.zone.org', 'A', '1.2.3.4''),
                    ('bar.zone.org', 'NS', ['ns1.zone.org', 'ns2.zone.org'])
            )

        Whenever records of a given type are updated, all records not included in the new RR set are deleted.  That is, if the ``A`` records are being updated, all current addresses must be included.
        '''

        raise NotImplementedError


__all__ += ['DnsZone']


class PublicDnsManagement(AsyncInjectable):

    '''
Update a DNS zone when :class:`NetworkLinks` gain a public IP address.  This can be attached to an existing injector if *attach_to* is included on construction.  More typically, this can be used as a Carthage modeling mixin.
    Typical usage in that mode::

        class some_enclave(Enclave, PublicDnsManagement):

            domain = "machines.example.com"
            add_provider(InjectionKey(DnsZone, name='machines.example.com', addressing='public'), some_dns_zone)

            class some_machine(MachineModel): ...

    Then, when `some_machine` gains an IP address, an `A` record will be created.

    '''

    async def public_ip_updated(self, target, **kwargs):
        link = target
        model = link.machine
        name = link.public_dns_name
        if name is None: name = model.name
        if not name: return     # public_dns_name = ''
        stem, sep, domain = name.partition('.')
        if not domain: return
        zone = await self.ainjector.get_instance_async(InjectionKey(
            DnsZone, name=domain, addressing='public', _ready=True, _optional=True))
        if zone is None:
            zone = await self.ainjector.get_instance_async(InjectionKey(DnsZone, role='public_zone', _ready=True, _optional=True))
        if zone is None and link.public_dns_name:
            logger.error(f'No public zone found for {name}')
        elif zone is None:
            logger.debug(f'No public zone found for {name}, but it may not be required')
            
        if not zone.contains(name):
            logger.error(f'Not setting DNS for {model}: {name} does not fall within {zone.name}')
        else:
            logger.debug(f'{name} is at {str(link.merged_v4_config.public_address)}')
            await zone.update_records((name, 'A', [str(link.merged_v4_config.public_address)]),
                                      ttl=30)

    def __init__(self, attach_to=None, **kwargs):
        super().__init__(**kwargs)
        if attach_to is None:
            attach_to = self.injector
        attach_to.add_event_listener(InjectionKey(NetworkLink), 'public_address', self.public_ip_updated)


__all__ += ['PublicDnsManagement']

@inject(ainjector=AsyncInjector)
async def update_dns_for(name, *,
                         public_records,
                         private_records,
                         ainjector, ttl=300,
                         ):
    '''
    Look up the zone for *name*, and update records within it.
    For public records,  ``InjectionKey(DnsZone, name=domain, addressing='public')`` is used.

    If a public zone is found, then ``InjectionKey(DnsZone, name=domain, addressing='private')`` is used for private records.  If a public zone is not found, then ``InjectionKey(DnsZone, name=domain)`` will also be accepted for private records.

    The zone keys used are dependent on what zones are registered in the injector hierarchy, **not** on what parameters are passed in.  That is, if a public zone exists, and only private records are passed in, private dns will only be updated if an explicit ``addressing='private'`` zone is provided by the injector.
    

    :param records: A sequence of (rrtype, values)

    Example Usage::

        await ainjector(update_dns_for,
            "www.foo.com", public_records=[('CNAME', 'foo.com')])

    '''
    head, sep, domain = name.partition('.')
    public_zone = await ainjector.get_instance_async(InjectionKey(DnsZone, name=domain, addressing='public', _optional=True))
    private_zone = await ainjector.get_instance_async(InjectionKey(DnsZone, name=domain, addressing='private', _optional=True))
    if (not public_zone) and (not private_zone):
        private_zone = await ainjector.get_instance_async(InjectionKey(DnsZone, name=domain, _optional=True))
    futures = []
    if public_records and public_zone:
        args = []
        for type, value in public_records:
            args.append((name, type, value))
        futures.append(asyncio.ensure_future(public_zone.update_records(*args, ttl=ttl)))
    elif public_records:
        logger.warning(f'No public zone for {name}')
    if private_zone and private_records:
        args = []
        for type, value in private_records:
            args.append((name, type, value))
        futures.append(asyncio.ensure_future(private_zone.update_records(*args, ttl=ttl)))

    if private_zone is None and public_zone is None:
        logger.warning(f'No DNS zone for {name}')
    if futures:
        await asyncio.gather(*futures)
        

__all__ += ['update_dns_for']



# Copyright (C) 2018, 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import carthage.dependency_injection

from .vm import Vm, VmFolder, VmTemplate
from .image import *
from .inventory import *
from .cluster import VmwareCluster
from .utils import wait_for_task
from .connection import VmwareConnection
from .network import DistributedPortgroup, vmware_trunk_key, DvSwitch
from .host import *
from .folder import *
from . import image

import carthage.vmware.network as network


@carthage.dependency_injection.inject(injector=carthage.Injector)
def enable_new_vmware_connection(injector):
    '''
    Indicate a point in the injector hierarchy where potentially a different connection is required.  Ifg only one connection is required in a layout this function is automatically called when the *carthage.vmware* plugin is enabled.  However, this function can be called to allow for selection of a new connection, datacenter, folder and distributed switch at some point in the injector hierarchy.
    '''
    from carthage.dependency_injection import partial_with_dependencies
    from ..utils import when_needed
    from . import inventory
    from .datacenter import VmwareDatacenter
    injector.add_provider(VmFolder)
    injector.add_provider(vmware_trunk_key, network._vmware_trunk)
    injector.add_provider(
        image.vm_datastore_key,
        partial_with_dependencies(
            image.produce_datastore_from_config,
            "vmware.datastore"))
    injector.add_provider(VmwareDatacenter)
    injector.add_provider(VmwareCluster)
    injector.add_provider(DvSwitch)
    injector.add_provider(VmwareConnection)
    injector.add_provider(
        image.image_datastore_key,
        partial_with_dependencies(
            image.produce_datastore_from_config,
            "vmware.image_datastore"))


@carthage.dependency_injection.inject(
    injector=carthage.dependency_injection.Injector)
def carthage_plugin(injector):
    from ..utils import when_needed
    from . import inventory
    from .datacenter import VmwareDatacenter
    injector.add_provider(inventory.custom_fields_key, inventory.default_custom_fields)
    injector.add_provider(DistributedPortgroup, allow_multiple=True)
    injector(enable_new_vmware_connection)



# Copyright (C) 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import contextlib
import logging
import os
import re
import tempfile
from .utils import import_resources_files
from pathlib import Path
from .dependency_injection import *
from .setup_tasks import *
from .config import ConfigLayout
from .image import ContainerImage, ContainerCustomization, ContainerVolume, ImageVolume
from .machine import customization_task
from . import sh

logger = logging.getLogger('carthage')

file_re = re.compile(r'file:/+(/[^/]+.*)')

__all__ = []


def bind_args_for_mirror(mirror):
    match = file_re.match(mirror)
    if match:
        return [f'--bind={match.group(1)}']
    return []


__all__ += ['bind_args_for_mirror']


@inject_autokwargs(config_layout=ConfigLayout)
class DebianContainerCustomizations(ContainerCustomization):

    description = "Set up Debian for Carthage"

    @setup_task("Turn on networkd")
    async def turn_on_networkd(self):
        root = Path(self.path)
        if not root.joinpath("usr/bin/resolvectl").exists():
            await self.container_command('apt', 'update')
            await self.container_command('apt', '-y', 'install', 'systemd-resolved')
        await self.container_command("systemctl", "enable", "systemd-networkd", "systemd-resolved")

    @setup_task("Install python and dbus")
    async def install_python(self):
        bind_args = bind_args_for_mirror(self.config_layout.debian.stage1_mirror)
        async with use_stage1_mirror(self):
            await self.container_command(*bind_args,
                                         "apt", "update")
            await self.container_command(*bind_args,
                                         "apt-get", "-y", "install", "python3", "dbus")


class DebianContainerImage(ContainerImage):

    mirror: str
    distribution: str
    include_security: bool
    name: str = "base-debian"

    def __init__(self, name: str = None,
                 mirror: str = None, distribution: str = None,
                 stage1_mirror: str = None,
                 include_security: bool = None,
                 **kwargs):
        if name is None:
            name = self.__class__.name
        super().__init__(name=name, **kwargs)
        # Make sure our config layout points to our injector so we can
        # change things with only local effect.
        self.config_layout = self.injector(ConfigLayout)
        self.mirror = self.config_layout.debian.mirror
        self.stage1_mirror = self.config_layout.debian.stage1_mirror
        self.distribution = self.config_layout.debian.distribution
        self.include_security = self.config_layout.debian.include_security
        self.debootstrap_options = self.config_layout.debian.debootstrap_options
        if mirror:
            self.mirror = mirror
            if not stage1_mirror:
                self.stage1_mirror = mirror
        if distribution:
            self.distribution = distribution
        if stage1_mirror:
            self.stage1_mirror = stage1_mirror
        if include_security is not None:
            self.include_security = include_security
        # Make sure we save the right items in case a customization ends up calling update_mirror.
        dc = self.config_layout.debian
        dc.mirror = self.mirror
        dc.distribution = self.distribution
        dc.stage1_mirror = self.stage1_mirror
        dc.include_security = self.include_security

    @setup_task("unpack using debootstrap")
    async def unpack_container_image(self):
        await sh.debootstrap('--include=openssh-server,ca-certificates',
                             *(self.debootstrap_options.split()),
                             self.distribution,
                             self.path, self.stage1_mirror,
                             _bg=True,
                             _bg_exc=False)
        path = Path(self.path)
        try:
            os.unlink(path / "etc/hostname")
        except FileNotFoundError:
            pass
        try:
            with path.joinpath("etc/ssh/sshd_config").open("at") as f:
                f.write("PasswordAuthentication no")
        except FileNotFoundError:
            pass


    @setup_task("Update mirror")
    def update_mirror(self):
        update_mirror(self.path, self.mirror, self.distribution, self.include_security)

    debian_customizations = customization_task(DebianContainerCustomizations)


__all__ += ['DebianContainerImage']


def update_mirror(path, mirror, distribution, include_security):
    etc_apt = Path(path) / "etc/apt"
    sources_list = etc_apt / "sources.list"
    if sources_list.exists():
        os.unlink(sources_list)
    debian_list = etc_apt / "sources.list.d/debian.list"
    os.makedirs(debian_list.parent, exist_ok=True)
    with debian_list.open("wt") as f:
        f.write(f'''
deb {mirror} {distribution} main contrib non-free
deb-src {mirror} {distribution} main contrib non-free
''')
        if include_security:
            f.write(f'''
deb {mirror} {distribution}-updates main contrib non-free
deb-src {mirror} {distribution}-updates main contrib non-free

deb https://security.debian.org/ {distribution}-security main
''')


@contextlib.asynccontextmanager
async def use_stage1_mirror(machine):
    debian = machine.config_layout.debian
    async with machine.filesystem_access() as path:
        try:
            update_mirror(path, debian.stage1_mirror, debian.distribution,
                          (debian.include_security and (debian.mirror == debian.stage1_mirror)))
            if machine.running:
                await machine.ssh("apt", "update",
                                  _bg=True, _bg_exc=False)
            else:
                await machine.container_command(*bind_args_for_mirror(debian.stage1_mirror),
                                                "apt", "update")
            yield
        finally:
            update_mirror(path, debian.mirror, debian.distribution, debian.include_security)
            try:
                if machine.running:
                    await machine.ssh("apt", "update",
                                      _bg=True, _bg_exc=False)
                else:
                    await machine.container_command(*bind_args_for_mirror(debian.mirror),
                                                    "apt", "update")
            except BaseException:
                logger.exception("Error cleaning up mirror")

__all__ += ['use_stage1_mirror']


def install_stage1_packages_task(packages):
    @setup_task(f'Install {packages} using stage 1 mirror')
    async def install_task(self):
        async with use_stage1_mirror(self):
            mirror = self.config_layout.debian.stage1_mirror
            await self.container_command(
                *bind_args_for_mirror(mirror),
                'apt', '-y',
                'install', *packages)
    return install_task


__all__ += ['install_stage1_packages_task']


@inject(ainjector=AsyncInjector, config=ConfigLayout)
async def debian_container_to_vm(
        volume: ContainerVolume,
        output: str,
        size: str,
        classes: str = None,
        *,
        config,
        image_volume_class=ImageVolume,
        ainjector):
    '''

    Use FAI to convert a container image into a VM image.  Carthage's FAI configuration currently requires that the host running Carthage (and probably the container being converted) use systemd-resolved for DNS.

    :param size: Size of resulting disk; for example ``8g``

    :param classes: The FAI classes to use.  If this starts with a "+", then add to the default classes.  The following classes are available:

        * SERIAL: enable serial console

        * DEFAULT: mandatory behavior

        * GROW: grow the root partition to fill the disk if the disk is expanded

        * GRUB_EFI: Install EFI version of grub.

        * CLOUD_INIT: Enable cloud-init


    '''
    from .container import logger

    def out_cb(data):
        data = data.strip()
        logger.debug("Image Creation: %s", data)

    async def unpack_callback(image_volume):
        nonlocal volume
        if isinstance(volume, type):
            # Allow usage specifying volume in a when_needed
            volume = await ainjector(volume)
        await volume.async_become_ready()
        os.makedirs(output_path.parent, exist_ok=True)
        with tempfile.TemporaryDirectory(dir=output_path.parent,
                                         prefix="container-to-vm-") as tmp_d:
            tmp = Path(tmp_d).absolute()
            await sh.tar(
                "-C", str(volume.path),
                "--xattrs",
                "--xattrs-include=*.*",
                "-czf",
                str(tmp / "base.tar.gz"),
                ".",
                _bg=True,
                _bg_exc=False)
            env = os.environ.copy()
            env['FAI_BASE'] = str(tmp / "base.tar.gz")
            await sh.fai_diskimage(
                '-S', size,
                '-s', str(fai_configspace),
                '-c', classes,
                str(tmp / "image.raw"),
                _env=env,
                _bg=True,
                _bg_exc=False,
                _encoding='UTF-8',
                _out=out_cb)
            os.rename(tmp / "image.raw", output_path)

    default_classes = "DEFAULT,CONTAINER2VM,GRUB_EFI"
    if classes is None:
        classes = default_classes
    elif classes[0] == "+":
        classes = default_classes + ',' + classes[1:]

    output_path = Path(config.vm_image_dir).joinpath(output)
    return await ainjector(image_volume_class, name=output_path.absolute(),
                           unpack=unpack_callback,
                           )

__all__ += ['debian_container_to_vm']

#: The Carthage FAI configuration space
fai_configspace = import_resources_files(__package__) / "resources/fai"



# Copyright (C) 2019, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import sys


def get_tb(e):
    if isinstance(e, BaseException):
        return e.__traceback__
    return e


def iter_tb(tb):
    while tb.tb_next:
        yield tb
        tb = tb.tb_next


def filter_before_here(e):
    tb = get_tb(e)
    frame = sys._getframe(1)
    for i in iter_tb(tb):
        if i.tb_frame is frame:
            if isinstance(e, BaseException):
                e.__traceback__ = i
            return tb
    return tb


def filter_chatty_modules(e, module_list, level=1):
    '''
    Filter chatty modules from an exception or traceback

    :param e: a :class:`Traceback` or :class:`BaseException` to be filtered in place

    :param module_list: A list of modules that are viewed as too chatty.  At most one consecutive entry from these modules will be retained.

    :param level: How many stack frames to look back for the caller.  Nothing before the caller will be filtered.  If *level* is None, the entire stack is filtered.

    This function first looks and finds the caller *level* levels above this call.

    #. The caller frame and all frames in the exception stack before the caller are left alone.

    #. Any frames that come from files whose modules are in *module_list* are filtered

    #. The filtering stops at the first frame not in *module_list*.

    '''
    tb = get_tb(e)
    if level is not None:
        caller = sys._getframe(level)
        caller_found = False
    else:
        caller_found = True
    module_filenos = frozenset(x.__file__ for x in module_list)
    while not caller_found and tb.tb_next is not None:
        if not caller_found:
            if caller is tb.tb_frame:
                caller_found = True
            else:
                tb = tb.tb_next
        else:  # caller already found
            break
    if caller_found:
        while (tb.tb_next is not None) and tb.tb_next.tb_next is not None:
            if tb.tb_next.tb_frame.f_code.co_filename in module_filenos \
                    and tb.tb_next.tb_next.tb_frame.f_code.co_filename in module_filenos:
                tb.tb_next = tb.tb_next.tb_next
            else:
                tb = tb.tb_next


__all__ = ('filter_before_here', 'filter_chatty_modules')



# Copyright (C)  2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.
import json
import os
import pytest
import shutil
from pathlib import Path
from carthage.podman import *
from carthage.oci import oci_container_image, OciExposedPort, OciMount
from carthage.ansible import *
from carthage.container import container_image
from carthage.modeling import *
from carthage.image import SshAuthorizedKeyCustomizations
from carthage.ssh import SshKey
from carthage import *
from carthage.machine import FilesystemCustomization
import carthage
from carthage.pytest import *

state_dir = Path(__file__).parent.joinpath("test_state")

@pytest.fixture(scope='session')
def enable_podman():
    import carthage.plugins
    base_injector(carthage.plugins.load_plugin, 'carthage.podman')

@pytest.fixture()
def ainjector(ainjector, enable_podman):
    ainjector = ainjector.claim("test_setup.py")
    config = ainjector.injector(carthage.ConfigLayout)
    config.state_dir = state_dir
    state_dir.mkdir(parents=True, exist_ok=True)
    yield ainjector
    shutil.rmtree(state_dir, ignore_errors=True)

@pytest.fixture()
def layout_fixture(ainjector):
    loop = ainjector.loop
    ainjector.add_provider(podman_layout)
    layout = loop.run_until_complete(ainjector.get_instance_async(CarthageLayout))
    yield layout
    try: loop.run_until_complete(
            layout.podman_net.instantiated.delete(force=True))
    except Exception: pass


class podman_layout(CarthageLayout):
    layout_name = 'podman'

    add_provider(machine_implementation_key, dependency_quote(PodmanContainer))
    add_provider(oci_container_image, 'debian:latest')
    #add_provider(ansible_log, "/tmp/ansible.log")

    oci_interactive = True

    @provides('podman_net')
    class podman_net(NetworkModel):
        v4_config = V4Config(
            network='10.66.0.0/24',
            pool=('10.66.0.5', '10.66.0.20'))

        instantiated = injector_access(carthage.podman.PodmanNetwork)
    class FromScratchDebian(PodmanFromScratchImage):
        oci_image_cmd = 'bash'
        oci_image_tag = 'localhost/from_scratch_debian'

    class DebianWithAuthorizedKeys(PodmanImage):
        oci_image_tag = 'localhost/authorized-debian:latest'
        authorized_keys = image_layer_task(SshAuthorizedKeyCustomizations)

    class ImageModelCustomizations(PodmanImageModel):

        oci_image_tag = 'localhost/carthage_podman:image_model'

        class fs_cust(FilesystemCustomization):

            @setup_task("Test filesystem customization in image model")
            def test_fscust(self): pass

        class container_cust(ContainerCustomization):

            @setup_task("Test container customization")
            def test_container_cust(self): pass
            
    class foo(MachineModel):

        name = 'foo.com'

    class ssh_test(MachineModel):
        name = 'ssh-test.foo.com'
        ip_address = '127.0.0.1'

        add_provider(OciExposedPort(22))

    class mount_test(MachineModel):
        add_provider(OciMount(
            mount_type='bind',
            destination='/host',
            source='/',
        ))

    class ansible_test(MachineModel):

        class cust(FilesystemCustomization):

            @setup_task("Install Ansible")
            async def install_ansible(self):
                await self.run_command('apt', 'update')
                await self.run_command('apt', '-y', 'install', 'ansible')

            do_roles = ansible_role_task(os.path.dirname(__file__) + "/resources/test_ansible_role")

    class stamps_discarded(MachineModel):

        #A machine to confirm that stamps are ignored after a machine is deleted
        task_called = False
        class cust(FilesystemCustomization):

            @setup_task("Set variable")
            def set_variable(self):
                self.host.model.task_called = True

    class TrueImage(ContainerfileImageModel):

        oci_image_tag = 'localhost/true:latest'
        container_context = 'resources/true_container'

    class DynamicContainerFileImage(ContainerfileImage):
        oci_image_tag = 'localhost/dynamic:latest'
        container_context = 'resources/dynamic_container'

        @setup_task('Create dynamic script')
        def create_dynamic_script(self):
            self.output_path.joinpath('script').write_text('''\
#!/bin/sh
exit 0
            ''')
            
            
    class true_machine(MachineModel):
        add_provider(oci_container_image, injector_access(TrueImage))

        name = 'true-machine'

    class dynamic_machine(MachineModel):
        add_provider(oci_container_image, injector_access(DynamicContainerFileImage))

        name = 'dynamic-machine'
        
    class pod_group(ModelGroup):
        add_provider(OciExposedPort(22))

        @provides(InjectionKey(PodmanPod))
        class pod(PodmanPod):
            name = 'carthage-test-pod'

        class pod_member(MachineModel):
            pass

    class networked_container(MachineModel):

        class config(NetworkConfigModel):
            add('eth0',
                mac=None,
                net=podman_net)


    class net_pod(PodmanPodModel):

        name = 'net_test'
        class config(NetworkConfigModel):
            add('eth0', mac=None, net=podman_net)

        class container(MachineModel):
            pass

@async_test
async def test_podman_create(ainjector):
    l = await ainjector(podman_layout)
    ainjector = l.ainjector
    machine = l.foo.machine
    await machine.async_become_ready()
    assert await machine.find()
    machine.stop_timeout = 1
    async with machine.machine_running(ssh_online=False):
        assert await machine.is_machine_running()
    await machine.delete()
    assert not await machine.find()


@async_test
async def test_container_exec(ainjector):
    l = await ainjector(podman_layout)
    ainjector = l.ainjector
    machine = l.foo.machine
    try:
        await machine.async_become_ready()
        machine.stop_timeout = 1
        async with machine.machine_running(ssh_online=False):
            assert 'root' in str(await machine.container_exec('ls'))
    finally:
        await machine.delete()


@async_test
async def test_container_ssh(ainjector):
    l = await ainjector(podman_layout)
    ainjector = l.ainjector
    machine = l.ssh_test.machine
    await ainjector.get_instance_async(SshKey)
    try:
        await machine.async_become_ready()
        machine.stop_timeout = 1
        async with machine.machine_running(ssh_online=False):
            await machine.container_exec('apt', 'update')
            await machine.container_exec(
                'apt', '-y', '--no-install-recommends', 'install', 'openssh-server')
            await machine.apply_customization(SshAuthorizedKeyCustomizations)
            await machine.container_exec('mkdir', '/run/sshd')
            await machine.container_exec('/usr/sbin/sshd')
            await machine.ssh_online()
    finally:
        await machine.delete()


@async_test
async def test_podman_image(ainjector):
    l = await ainjector(podman_layout)
    ainjector = l.ainjector
    await l.DebianWithAuthorizedKeys.async_become_ready()


@async_test
async def test_podman_mount(ainjector):
    l = await ainjector(podman_layout)
    ainjector = l.ainjector
    machine = l.mount_test.machine
    assert machine.mounts
    try:
        machine.stop_timeout = 0
        await machine.async_become_ready()
        async with machine.machine_running(ssh_online=False):
            await machine.container_exec('ls', '/host/etc')
    finally:
        await machine.delete()


@pytest.mark.no_rootless
@async_test
async def test_from_scratch_image(test_ainjector):
    l = await test_ainjector(podman_layout)
    ainjector = l.ainjector
    config = await test_ainjector(ConfigLayout)
    config.delete_volumes = False
    ainjector.add_provider(podman_image_volume_key, injector_access(container_image))
    await l.FromScratchDebian.async_become_ready()


@async_test
async def test_podman_ansible(ainjector):
    l = await ainjector(podman_layout)
    ainjector = l.ainjector
    machine = l.ansible_test.machine
    try:
        await machine.async_become_ready()
    finally:
        await machine.delete()


@async_test
async def test_podman_pod(ainjector):
    l = await ainjector(podman_layout)
    ainjector = l.ainjector
    pg = l.pod_group
    machine = pg.pod_member.machine
    try:
        await machine.async_become_ready()
    finally:
        try:
            await pg.pod.delete(force=True)
        except Exception:
            pass

@async_test
async def test_stamps_ignored(ainjector):
    l = await ainjector(podman_layout)
    ainjector = l.ainjector
    assert l.stamps_discarded.task_called is False
    await l.stamps_discarded.machine.async_become_ready()
    assert l.stamps_discarded.task_called is True
    await l.stamps_discarded.machine.delete()
    # Podman doesn't work very well if you do something to an instance after delete
    # So we instantiate a second instance directly with the class.
    # Note that we're instantiating to ready since we call the injector directly
    stamps_discarded_2 = await ainjector(podman_layout.stamps_discarded)
    try:
        await stamps_discarded_2.machine.async_become_ready()
        assert stamps_discarded_2.task_called
    finally:
        try: await stamps_discarded_2.machine.delete()
        except Exception: pass

@async_test
async def test_containerfile_image(ainjector):
    l = await ainjector(podman_layout)
    ainjector = l.ainjector
    try:
        await l.true_machine.machine.async_become_ready()
    finally:
        try: await l.true_machine.machine.delete()
        except Exception: pass
@async_test
async def test_dynamic_containerfile_image(layout_fixture):
    l = layout_fixture
    ainjector = l.ainjector
    try:
        await l.dynamic_machine.machine.async_become_ready()
    finally:
        try: await l.dynamic_machine.machine.delete()
        except Exception: pass

@async_test
async def test_podman_container_network(layout_fixture):
    "Test networking on a single container"
    layout = layout_fixture
    try:
        await layout.networked_container.machine.async_become_ready()
    except InjectionFailed as e:
        if isinstance(e.__cause__,NotImplementedError):
            pytest.xfail("Podman too old")
    finally:
        try: await layout.networked_container.machine.delete()
        except Exception: pass

@async_test
async def test_podman_pod_network(layout_fixture):
    "Test networking in a pod"
    layout = layout_fixture
    try:
        try: await layout.net_pod.container.machine.async_become_ready()
        except InjectionFailed as e:
            if isinstance(e.__cause__, NotImplementedError):
                pytest.xfail('Podman is too old')
            raise
        async with layout.net_pod.container.machine.machine_running():
            await layout.net_pod.container.machine.container_exec('apt', 'update')
            machine = layout.net_pod.container.machine
            await machine.container_exec('apt', '-y', 'install', 'iproute2')
            result = await machine.container_exec('ip', 'addr', 'show')
            address =machine.network_links['eth0'].merged_v4_config.address
            assert address
            assert str(address) in str(result.stdout, 'utf-8')
    finally:
        try: await layout.net_pod.pod.delete(force=True)
        except Exception: pass

@async_test
async def test_podman_image_model(layout_fixture):
    await layout_fixture.ImageModelCustomizations.build_image()
    



# Copyright (C)  2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import asyncio
import contextlib
import datetime
import json
import logging
from pathlib import Path
import tempfile
import shutil
import dateutil.parser
import carthage.machine
from carthage.dependency_injection import *
from .. import sh, ConfigLayout
from ..machine import AbstractMachineModel, Machine
from ..utils import memoproperty
from ..network import TechnologySpecificNetwork, Network, V4Config, this_network, NetworkConfig
from ..oci import *
from ..setup_tasks import setup_task, SetupTaskMixin, TaskWrapperBase, SkipSetupTask


logger = logging.getLogger('carthage.podman')


def podman_port_option(p: OciExposedPort):
    res = f'-p{p.host_ip}:{p.host_port}:{p.container_port}'
    if p.proto != 'tcp':
        res += f'/{p.proto}'
    return res


def podman_mount_option(injector: Injector, m: OciMount):
    res = f'--mount=type={m.mount_type}'
    if m.source:
        res += f',source={m.source_resolved(injector)}'
    if m.destination:
        res += f',destination={m.destination}'
    else:
        raise TypeError('destination is required')
    if m.options:
        res += f',{m.options}'
    return res


__all__ = []


class PodmanContainerHost(AsyncInjectable):

    @memoproperty
    def podman_log(self):
        return self.injector.get_instance(InjectionKey("podman_log", _optional=True))
    
    def podman(self, *args,
               _bg=True, _bg_exc=True):
        raise NotImplementedError

    async def filesystem_access(self, container):
        raise NotImplementedError

    async def tar_volume_context(self, volume):
        '''
        An asynchronous context manager that tars up a volume and provides a path to that tar file usable in ``podman import``.  Typical usage::

            async with container_host.tar_volume_context(container_image) as path:
                await container_host.podman('import', path)

        On local systems this manages temporary directories.  For remote container hosts, this manages to get the tar file to the remote system and clean up later.
        '''
        raise NotImplementedError


class LocalPodmanContainerHost(PodmanContainerHost):

        
    
    @contextlib.asynccontextmanager
    async def filesystem_access(self, container):
        result = await self.podman(
            'container', 'mount',
            container,
            _bg=True, _bg_exc=False, _log=False)
        try:
            path = str(result).strip()
            yield Path(path)
        finally:
            pass  # Perhaps we should unmount, but we'd need a refcount to do that.

    def podman(self, *args,
               _bg=True, _bg_exc=False, _log=True):
        options = {}
        if _log and self.podman_log:
            options['_out']=str(self.podman_log)
            options['_err_to_out'] = True
        return sh.podman(
            *args,
            _bg=_bg, _bg_exc=_bg_exc,
            _encoding='utf-8',
            **options)

    @contextlib.asynccontextmanager
    async def tar_volume_context(self, volume):
        assert hasattr(volume, 'path')
        with tempfile.TemporaryDirectory() as path_raw:
            path = Path(path_raw)
            await sh.tar(
                "-C", str(volume.path),
                "--xattrs",
                "--xattrs-include=*.*",
                "-czf",
                str(path / "container.tar.gz"),
                ".",
                _bg=True,
                _bg_exc=False)
            yield path / 'container.tar.gz'

class PodmanNetworkMixin:

    async def _setup_networks(self):
        for l in self.network_links.values():
            if l.local_type: continue
            await l.instantiate(PodmanNetwork)

    async def _network_options(self):
        await self._setup_networks()
        options = []
        for l in self.network_links.values():
            if l.local_type: continue
            l.net.assign_addresses()
            options.extend(['--network', l.net_instance.link_options(l)])
        return options

    async def _container_network_options(self):
        # Like network_options but handles network_namespace_key
        try:
            network_namespace = await self.ainjector.get_instance_async(carthage.machine.network_namespace_key)
        except KeyError: network_namespace = False
        if self.pod and self.network_links and network_namespace is False:
            logger.error(f"{self.name} will not join network of {self.pod}; if this is intended then add_provider(network_namespace_key, dependency_quote(None) in {self.name}'s model.  If it is not, then set network_namespace_key to {self.pod}")
        if network_namespace:
            if isinstance(network_namespace, PodmanContainer):
                await network_namespace.find_or_create()
                return ['--network', 'container:'+network_namespace.id]
            return []           # Joining a pod
        return await self._network_options() # Creating our own namespace
           
    async def resolve_networking(self, force:bool = False):
        await super().resolve_networking(force=force)
        for net in set(map( lambda l:l.net, self.network_links.values())):
            net.assign_addresses()
            
@inject(
    podman_pod_options=InjectionKey('podman_pod_options', _optional=NotPresent),
)
class PodmanPod(OciPod, PodmanNetworkMixin, carthage.machine.NetworkedModel):

    #: A list of extra options to pass to pod create
    podman_pod_options = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.injector.add_provider(InjectionKey(PodmanPod), dependency_quote(self))
        self.network_links = {}

    async def find(self):
        if self.id:
            inspect_arg = self.id
        else:
            inspect_arg = self.name
        try:
            result = await self.podman(
                'pod', 'inspect', inspect_arg, _log=False)
        except sh.ErrorReturnCode:
            return False
        pod_info = json.loads(str(result.stdout, 'utf-8'))
        self.pod_info = pod_info
        return dateutil.parser.isoparse(pod_info['Created']).timestamp()

    async def do_create(self):
        await self.resolve_networking()
        options = await self._network_options()
        for p in self.exposed_ports:
            options.append(podman_port_option(p))
        options.extend(self.podman_pod_options)
        await self.podman(
            'pod', 'create',
            *options,
            '--name=' + self.name)

    async def delete(self, force=False):
        force_options = []
        if force:
            force_options.append('--force')
        await self.podman(
            'pod', 'rm',
            *force_options,
            self.name)

    @memoproperty
    def podman(self):
        return self.injector(LocalPodmanContainerHost).podman


__all__ += ['PodmanPod']

   

@inject_autokwargs(
    oci_container_image=InjectionKey(oci_container_image, _optional=NotPresent),
    podman_restart=InjectionKey('podman_restart', _optional=NotPresent),
    pod=InjectionKey(PodmanPod, _optional=True),
    podman_options=InjectionKey('podman_options', _optional=NotPresent),
)
class PodmanContainer(PodmanNetworkMixin, Machine, OciContainer):

    '''
An OCI container implemented using ``podman``.  While it is possible to set up a container to be accessible via ssh and to meet all the interfaces of :class:`~carthage.machine.SshMixin`, this is relatively uncommon.  Such containers often have an entry point that is not a full init, and only run one service or program.  Typically :meth:`container_exec` is used to execute an additional command in the scope of a container rather than using :meth:`ssh`.
    '''

    #: Timeout in seconds to wait when stopping a container
    stop_timeout = 10
    machine_running_ssh_online = False
    rsync_uses_filesystem_access = True

    #: restart containers (no, always, on-failure)
    podman_restart = 'no'

    #:Extra options (as a list) to be passed into podman create
    podman_options = []

    @memoproperty
    def ssh_options(self):
        if not hasattr(self, 'ssh_port'):
            raise ValueError('Set ssh_port before ssh')
        return (
            *super().ssh_options,
            f'-p{self.ssh_port}')

    #:The port on which to connect to for ssh
    ssh_port: int

    @memoproperty
    def ansible_inventory_overrides(self):
        return dict(
            ansible_connection='containers.podman.podman',
            ansible_pipelining=False,
            ansible_host=self.full_name,
        )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._operation_lock = asyncio.Lock()

    @memoproperty
    def podman(self):
        return self.container_host.podman

    @memoproperty
    def container_host(self):
        return self.injector(LocalPodmanContainerHost)

    async def find(self):
        try:
            result = await self.podman(
                'container', 'inspect', self.full_name,
                _bg=True, _bg_exc=False, _log=False)
        except sh.ErrorReturnCode:
            return False
        containers = json.loads(str(result))
        self.container_info = containers[0]
        ports = self.container_info['NetworkSettings']['Ports']
        if not hasattr(self, 'ssh_port') and '22/tcp' in ports:
            self.ssh_port = ports['22/tcp'][0]['HostPort']
        self.id = self.container_info['Id']
        self.running = self.container_info['State']['Running']
        try:
            return dateutil.parser.isoparse(containers[0]['Created']).timestamp()
        except Exception as e:
            raise ValueError(f'Invalid ISO string: {self.container_info["Created"]}')

            
    async def do_create(self):
        image = self.oci_container_image
        if isinstance(image, OciImage):
            await image.async_become_ready()
            image = image.oci_image_tag
        if self.pod:
            await self.pod.async_become_ready()
        command_options = []
        if self.oci_command:
            command_options = list(self.oci_command)
        network_options = await self._container_network_options()
        await self.podman(
            'container', 'create',
            f'--name={self.full_name}',
            *self._podman_create_options(),
            *network_options,
            image,
            *command_options,
            _bg=True, _bg_exc=False)

    def _podman_create_options(self):
        options = []
        options.append('--restart=' + self.podman_restart)
        if self.oci_interactive:
            options.append('-i')
        if self.oci_tty:
            options.append('-t')
        for k, v in self.injector.filter_instantiate(
                OciEnviron, lambda k: 'name' in k.constraints and k.constraints.get('scope', 'all') in ('all', 'container')):
            options.append('-e' + v.assignment)
        if not self.pod:
            for p in self.exposed_ports:
                options.append(podman_port_option(p))
        else:  # there is a pod
            options.append('--pod=' + self.pod.name)
        for m in self.mounts:
            options.append(podman_mount_option(self.injector, m))
        options.extend(self.podman_options)
        return options

    async def delete(self, force=True, volumes=True):
        force_args = []
        if force:
            force_args.append('--force')
        if volumes:
            force_args.append('--volumes')
        await self.podman(
            'container', 'rm',
            *force_args, self.full_name,
            _bg=True, _bg_exc=False)

    async def is_machine_running(self):
        assert await self.find(), "Container does not exist"
        self.running = self.container_info['State']['Running']
        return self.running

    async def start_machine(self):
        async with self._operation_lock:
            await self.is_machine_running()
            if self.running:
                return
            await self.start_dependencies()
            await super().start_machine()
            logger.info(f'Starting {self.full_name}')
            await self.podman(
                'container', 'start', self.full_name,
                _bg=True, _bg_exc=False)
        await self.is_machine_running()

    async def stop_machine(self):
        async with self._operation_lock:
            if not self.running:
                return
            await self.podman(
                'container', 'stop',
                f'-t{self.stop_timeout}',
                self.full_name,
                _bg=True, _bg_exc=False)
            self.running = False
            await super().stop_machine()

    def container_exec(self, *args):
        '''
'Execute a command in a running container and return stdout.  This function intentionally has a differentname than :meth:`carthage.container.Container.container_command` because that method does not expect the container to be running.
'''
        result = self.podman(
            'container', 'exec',
            self.full_name,
            *args,
            _log=False)
        return result

    #: An alias to be more compatible with :class:`carthage.container.Container`
    shell = container_exec

    def _apply_to_filesystem_customization(self, customization):
        @contextlib.asynccontextmanager
        async def customization_context():
            async with self.machine_running(ssh_online=False), self.filesystem_access() as path:
                customization.path = path
                yield
            return
        customization.customization_context = customization_context()
        customization.run_command = self.container_exec

    def filesystem_access(self):
        return self.container_host.filesystem_access(self.full_name)

    def __repr__(self):
        try:
            host = repr(self.container_host)
        except Exception:
            host = "repr failed"
        return f'<{self.__class__.__name__} {self.name} on {host}>'

    @memoproperty
    def stamp_path(self):
        state_dir = Path(self.config_layout.state_dir)
        result = state_dir.joinpath("podman", self.name)
        result.mkdir(exist_ok=True, parents=True)
        return result


    def check_stamp(self, stamp, raise_on_error=False):
        mtime, text = super().check_stamp(stamp, raise_on_error)
        creation = getattr(self, '_find_result', None)
        if creation and mtime < creation:
            return False, "" #stamp predates container creation
        return mtime, text

__all__ += ['PodmanContainer']


class PodmanImageBuilderContainer(PodmanContainer):

    oci_command = ['sleep', '3600']
    stop_timeout = 1

    def _apply_to_container_customization(self, customization):
        @contextlib.asynccontextmanager
        async def customization_context():
            async with self.machine_running(ssh_online=False), self.filesystem_access() as path:
                customization.path = path
                customization.container_command = self.container_exec
                customization.running = False # It is technically running, but not with any services
                yield
            return
        customization.customization_context = customization_context()

    @memoproperty
    def model(self):
        from .modeling import PodmanImageModel
        res =  self.injector.get_instance(InjectionKey(PodmanImageModel, _ready=False, _optional=True))
        if res is None: raise AttributeError
        return res

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.injector.add_provider(InjectionKey(NetworkConfig), dependency_quote(None))
@inject_autokwargs(
    base_image=InjectionKey(oci_container_image, _optional=NotPresent),
)
class PodmanImage(OciImage, SetupTaskMixin):

    '''
    Represents an OCI container image and provides facilities for building the image.

    :class:`customizations <carthage.machine.BaseCustomization>` can be turned into image layers using the :func:`image_layer_customization` function.  Note that :func:`setup_tasks <setup_task>` are only run when images are actually built.  By default, the image is only built if it does not exist, although see :meth:`should_build` to override.

    '''

    last_layer = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.layer_number = 1
        self.injector.add_provider(InjectionKey('podman_log'), self.stamp_path/'podman.log')

    async def pull_base_image(self):
        if isinstance(self.base_image, OciImage):
            await self.base_image.async_become_ready()
            base_image = self.base_image.oci_image_tag
        else:
            base_image = self.base_image
        if not base_image.startswith('localhost/'):
            await self.podman(
                'pull', base_image,
            )
        inspect_result = await self.podman(
            'image', 'inspect',
            base_image, _log=False)
        image_info = json.loads(str(inspect_result))[0]
        self.parse_base_image_info(image_info)

    def parse_base_image_info(self, image_info):
        config = image_info['Config']
        if not self.oci_image_command and 'Cmd' in config:
            self.oci_image_command = config['Cmd']
        if not self.oci_image_entry_point and 'EntryPoint' in config:
            self.oci_image_entry_point = config['EntryPoint']
        self.base_image_info = image_info
        self.last_layer = self.base_image_info['Id']

    async def find(self):
        if self.id:
            to_find = self.id
            self.oci_read_only = True
        else:
            to_find = self.oci_image_tag
        try:
            result = await self.podman(
                'image', 'inspect', to_find,
                _log=False)
        except sh.ErrorReturnCode:
            return False
        info = json.loads(str(result))[0]
        self.id = info['Id']
        self.image_info = info
        return dateutil.parser.isoparse(info['Created']).timestamp()

    async def should_build(self):
        '''If the image exists, this is called.  If it returns True, then the image will be rebuilt even though it exists.  If a caller wants to force a rebuild, it is better to call :meth:`build_image` than to patch this method.
        '''
        return False

    @contextlib.asynccontextmanager
    async def image_layer_context(self, commit_message=""):
        '''
        Generate a container to produce  a new image layer:

        * The image of the container will be either *self.last_layer* or *self.base_image* if *last_layer* is not set.

        * The container will be a :class:`PodmanImageBuilderContainer`, and as such will simply pause when started so that :meth:`container_exec` can be used to run commands in the container.

        Usage::

            async with self.image_layer_context() as layer_container:
                # Apply customizations/run commands in layer_container
            #Now, self.last_layer is the image ID of the new layer
        '''
        def container_delete(future):
            try:
                future.result()
            except Exception as e:
                logger.error('Error deleting %s: %s', layer_container, str(e))

        base_image = self.last_layer or self.base_image
        layer_container = await self.ainjector(
            PodmanImageBuilderContainer,
            oci_container_image=base_image,
            name=f'carthage-image-build-{id(self)}-l{self.layer_number}',
        )
        self.layer_number += 1
        try:
            await layer_container.start_machine()
            yield layer_container
            await self.commit_container(layer_container, commit_message)
        finally:
            delete_task = asyncio.get_event_loop().create_task(layer_container.delete())
            delete_task.add_done_callback(container_delete)

    def _commit_options(self):
        entrypoint = None
        cmd = None
        if self.oci_image_entry_point:
            entrypoint = json.dumps(self.oci_image_entry_point)
        if self.oci_image_command:
            cmd = json.dumps(self.oci_image_command)
        options = []
        if cmd:
            options.append('--change=CMD ' + cmd)
        if entrypoint:
            options.append('--change=ENTRYPOINT ' + entrypoint)
        for k, v in self.injector.filter_instantiate(
                OciEnviron, lambda k: 'name' in k.constraints and k.constraints.get('scope', 'all') in ('all','image')):
            options.append('--change=ENV '+v.assignment)
        return options

    async def commit_container(self, container, commit_message):
        options = self._commit_options()
        if self.oci_image_author:
            options.append('--author=' + self.oci_image_author)
        if commit_message:
            options.append('-fdocker')
            options.append('--message=' + commit_message)
        # options must be quoted if it's going through ssh or something that can split args on space
        commit_result = await self.podman(
            'container', 'commit',
            *options,
            container.id, _log=False)
        self.last_layer = str(commit_result.stdout, 'utf-8').strip()

    async def tag_last_layer(self):
        assert self.last_layer
        await self.podman(
            'image', 'tag',
            self.last_layer, self.oci_image_tag)

    async def find_or_create(self):
        '''See if image exists otherwise rebuild the image.
        Note that this is not a :func:`setup_task` even though it is in the parent.  This is always run from :meth:`async_ready`
        '''
        if await self.find():
            if not await self.should_build():
                return
        return await self.build_image()

    async def build_image(self):
        await self.pull_base_image()
        # You might think that context for run_setup_tasks should be
        # self.image_layer_context().  If it worked that way, then
        # everything would end up in a single layer.  Instead, use
        # image_layer_task for wrapping customizations and explicitly
        # call image_layer_context in setup_tasks.
        await self.run_setup_tasks()
        if not self.last_layer:
            logger.warn('%s failed to generate any image layers', self)
            return
        await self.tag_last_layer()

    async def async_ready(self):
        await self.find_or_create()
        return await AsyncInjectable.async_ready(self)

    @memoproperty
    def podman(self):
        return self.container_host.podman

    @memoproperty
    def container_host(self):
        return self.injector(LocalPodmanContainerHost)

    @memoproperty
    def stamp_path(self):
        config = self.injector(ConfigLayout)
        path = Path(config.output_dir)/"podman_image"/self.oci_image_tag.replace('/','_')
        path.mkdir(exist_ok=True, parents=True)
        return path
    
                            

__all__ += ['PodmanImage']
podman_image_volume_key = InjectionKey('carthage.podman/image_volume')


@inject(base_image=None)
@inject_autokwargs(
    image_volume=podman_image_volume_key,
)
class PodmanFromScratchImage(PodmanImage):

    async def pull_base_image(self):
        await self.image_volume.async_become_ready()
        async with self.container_host.tar_volume_context(self.image_volume) as tar_path:
            result = await self.podman(
                'image', 'import',
                *self._commit_options(),
                tar_path,
                _log=False)
        id = str(result.stdout, 'utf-8').strip()
        inspect_result = await self.podman(
            'image', 'inspect',
            id, _log=False)
        image_info = json.loads(str(inspect_result))[0]
        self.last_layer = id
        self.parse_base_image_info(image_info)


__all__ += ['PodmanFromScratchImage', 'podman_image_volume_key']


def image_layer_task(customization):
    '''Wrap a :class:`~carthage.machine.BaseCustomization` as a layer in a :class:`PodmanImage`.
    '''
    return ImageLayerTask(customization)

__all__ += ['image_layer_task']

class ImageLayerTask(TaskWrapperBase):

    customization: BaseCustomization
    description: str

    def __init__(self, customization, **kwargs):
        self.customization = customization
        description = getattr(customization, 'description',
                                   'Image Layer: ' + customization.__name__)
        super().__init__(description=description, **kwargs)

    async def func(self, image:PodmanImage):
        async with image.image_layer_context(self.description) as container:
            await container.apply_customization(self.customization)

    async def check_completed_func(self, image):
        # We always want to rerun images
        return False

    @memoproperty
    def stamp(self):
        #Overridden if an image_layer_task is assigned as a class property
        return self.customization.__name__
    
    
@inject_autokwargs(
    config_layout=ConfigLayout,
    podman_options=InjectionKey('podman_options', _optional=NotPresent),
    )
class ContainerfileImage(OciImage):

    '''
    Build an image using ``podman build`` from a context directory with a ``Containerfile``.

    :param container_context: A directory with a Containerfile and potentially other files used by the Containerfile.  This can be specified either in a call to the constructor or in a subclass definition.  In the constructor, this is resolved relative to the current directory.  In a subclass, this is resolved relative to the package (or module) in which the class is defined.

    This class does respect :class:`OciMount` and :class:`OciEnviron` in the injector hierarchy.

    '''

    #: Options to pass to podman
    podman_options = tuple()

    def __init__(self, container_context=None, **kwargs):
        if container_context: self.container_context = container_context
        else:
            if not hasattr(self, 'container_context'):
                raise TypeError('container_context must be set on the class or in the constructor')
            try:
                import sys
                module = sys.modules[self.__class__.__module__]
            except Exception as e:
                module = None
                warnings.warn(f'Unable to find module for {self.__class__.__qualname__}: {e}')
            if module:
                try: path = Path(module.__path__[0])
                except Exception:
                    path = Path(module.__file__).parent
                self.source_container_context = self.container_context = path/self.container_context
        super().__init__(**kwargs)
        if len(self.setup_tasks) > 2:
            # More than just find_or_create and copy_context_if_needed
            self.setup_tasks.sort(key=lambda t: 1 if t.func == OciManaged.find_or_create.func else 0)
            self.container_context = self.output_path
        self.injector.add_provider(InjectionKey("podman_log"), self.stamp_path/'podman.log')
                                   
            

    @memoproperty
    def container_host(self):
        return self.injector(LocalPodmanContainerHost)

    @memoproperty
    def output_path(self):
        path = Path(self.config_layout.output_dir)/'podman_image'
        tag = self.oci_image_tag.replace('/', '_')
        path /= tag
        path.mkdir(exist_ok=True, parents=True)
        return path

    stamp_path = output_path

    @setup_task("Copy Context if Needed", order=10)
    async def copy_context_if_needed(self):
        if len(self.setup_tasks) > 2:
            #More than just this task and find_or_create
            logger.info('copying container context for %s image', self.oci_image_tag)
            shutil.rmtree(self.output_path)
            shutil.copytree(self.source_container_context, self.output_path, symlinks=True)
        else:
            raise SkipSetupTask

    @copy_context_if_needed.invalidator()
    def copy_context_if_needed(self, last_run):
        source_mtime = self.container_context_mtime(self.source_container_context)
        if source_mtime > last_run: return False
        return False
    
        
    async def do_create(self):
        options = await self._build_options()
        return await self.container_host.podman(
            'build',
            '--annotation', 'com.hadronindustries.carthage.image_mtime='+ \
            datetime.datetime.fromtimestamp(
                self.container_context_mtime(self.container_context),datetime.timezone.utc).isoformat(),
            '-t'+self.oci_image_tag,
            *options,
            self.container_context)

    async def find(self):
        try: inspect_result = await self.container_host.podman(
                'image', 'inspect',
                self.oci_image_tag, _log=False)
        except sh.ErrorReturnCode: return False
        inspect_json = json.loads(str(inspect_result.stdout, 'utf-8'))
        created = dateutil.parser.isoparse(inspect_json[0]['Created']).timestamp()
        hadron_mtime_str = inspect_json[0]['Annotations'].get('com.hadronindustries.carthage.image_mtime')
        if hadron_mtime_str:
            hadron_mtime = dateutil.parser.isoparse(hadron_mtime_str).timestamp()
            if self.container_context_mtime(self.container_context) > hadron_mtime+5: return False
            return hadron_mtime
        if self.container_context_mtime(self.container_context) > created:
            return False
        return created

    @staticmethod
    def container_context_mtime(container_context):
        context = Path(container_context)
        mtime = 0.0
        for p in context.iterdir():
            stat = p.stat()
            if stat.st_mtime >mtime: mtime = stat.st_mtime
        return mtime
    
    

    async def _build_options(self):
        options = []
        # Instantiate a container simply so we can ask it for volume, mount, and environment options.
        with instantiation_not_ready():
            container = await self.ainjector(PodmanContainer, name='image_options')
        for k, v in self.injector.filter_instantiate(
                OciEnviron, lambda k: 'name' in k.constraints and k.constraints.get('scope', 'all') in ('all','image')):
            options.append('--env')
            options.append(v.assignment)
        for m in container.mounts:
            options.append(podman_mount_option(self.injector, m))
        options.extend(self.podman_options)
        return options

__all__ += ['ContainerfileImage']

@inject_autokwargs(network=this_network)
class PodmanNetwork(TechnologySpecificNetwork, OciManaged):

    @memoproperty
    def container_host(self):
        return self.injector(LocalPodmanContainerHost)

    @property
    def podman(self):
        return self.container_host.podman
    
    async def find(self):
        try:
            inspect_result = await self.podman(
                'network', 'inspect', self.network.name, _log=False)
        except Exception: return False
        info = json.loads(str(inspect_result))[0]
        try:
            return dateutil.parser.isoparse(info['created']).timestamp()
        except (KeyError, ValueError):
            logger.error('Unable to understand network inspection result: %s', info)
            raise NotImplementedError('Podman too old')
        

    async def do_create(self):
        options = ['-d', 'bridge']
        v4_config = getattr(self.network, 'v4_config', None)
        if v4_config:
            if v4_config.network:
                options.extend([
                    '--subnet', str(v4_config.network)])
            if v4_config.gateway is False:
                options.append('--internal')
            elif v4_config.gateway and v4_config.gateway  is not True:
                options.extend([
                    '--gateway', str(v4_config.gateway)])
            if v4_config.dhcp:
                options.extend(['--ipam-driver=dhcp'])
        await self.podman(
            'network',
            'create', self.network.name,
            *options)

    async def delete(self, force=True):
        if force:
            force_options = ['--force']
        else: force_options = []
        await self.podman(
            'network', 'rm', *force_options,
            self.network.name)
        
    def link_options(self, link):
        def safe(s):
            assert ',' not in s
            assert '=' not in s
            return s
        v4_config = link.merged_v4_config
        options = ['interface_name='+safe(link.interface)]
        if v4_config.address:
            options.append('ip='+safe(str(v4_config.address)))
        if link.mac:
            options.append('mac='+safe(link.mac))
        if link.mtu:
            options.append('mtu='+safe(link.mtu))
        assert ':' not in self.network.name
        return safe(self.network.name)+':'+','.join(options)

__all__ += ['PodmanNetwork']



# Copyright (C) 2018, 2019, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import os.path
import pytest
from carthage.pytest import *
from carthage import base_injector, AsyncInjector, inject
from carthage import ConfigLayout
import carthage.ssh
'''This is a test for the pytest plugins, not for the carthage plugin mechanism'''

from machine_mock import Machine

resource_dir = os.path.dirname(__file__)


def test_test_parameters(test_parameters):
    return True


@async_test
async def test_async_test():
    return True


@async_test
async def test_async_test_with_loop(loop):
    assert loop


@async_test
@inject(config=ConfigLayout)
def test_carthage_injection(config, ainjector):
    assert config.delete_volumes == False


@async_test
@inject(ssh_key=carthage.ssh.SshKey)
async def test_mock_machine(ssh_key):
    m = Machine("Test machine")
    await ssh_key.rsync(os.path.join(resource_dir, 'test_carthage_plugin.py'),
                        m.rsync_path('/'))
    assert os.path.exists(os.path.join(
        m.path, "test_carthage_plugin.py"))


@async_test
@inject(ssh_key=carthage.ssh.SshKey)
async def test_carthage_controller(ssh_key, request, capsys):
    m = Machine('carthage-inner')
    await ssh_key.rsync(os.path.join(
        resource_dir,
        "inner_plugin_test.py"),
        m.rsync_path('/'))
    await ssh_key.rsync(
        os.path.join(resource_dir, "inner_conftest.py"),
        m.rsync_path("/conftest.py"))
    await subtest_controller(request, m, "inner_plugin_test.py")



# Copyright (C) 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import abc
from .dependency_injection import *
from .machine import Machine
from .utils import memoproperty
__all__ = []


class SystemDependency(abc.ABC, Injectable):

    '''Represents a dependency that may be required by a :class:`carthage.Machine` before a machine is started.  These dependencies may also be required by a :func:`carthage.setup_tasks.setup_task`; see :func:`depend_on()`.

    Note that typically instances of this class rather than subclasses are used as dependency providers on an Injector.  That way, :meth:`~carthage.Injector.get_instance` returns the instance without processing injected dependencies.  These dependencies are later processed when a method like :meth:`carthage.Machine.start_dependencies()` calls the :meth:`__call__()` method.


'''

    name: str
    @abc.abstractproperty
    def name(self): raise NotImplementedError

    @abc.abstractmethod
    def __call__(self, ainjector: AsyncInjector): raise NotImplementedError

    def default_instance_injection_key(self):
        return InjectionKey(SystemDependency, name=self.name, _globally_unique=True)

    @memoproperty
    def __provides_dependencies_for__(self):
        # This will make sure that dependencies in a modeling.InjectableModel are
        # added to the injector so they are found
        return [self.default_instance_injection_key()]

    def __repr__(self):
        return f'<SystemDependency name={self.name}'


__all__ += ['SystemDependency']


class MachineDependency(SystemDependency):

    ONLINE_DEFAULT = object()

    def __init__(self, m, *,
                 name=None,
                 online=ONLINE_DEFAULT):
        if isinstance(m, InjectionKey):
            self.key = InjectionKey(m, _ready=True)
        elif isinstance(m, str):
            self.key = InjectionKey(Machine, host=m, _ready=True)
        elif isinstance(m, Machine):
            self.key = InjectionKey(Machine, host=m.name, _ready=True)
        else:
            raise ValueError(f'unsupported value for MachineDependency: {m}')
        self.online = online
        if name:
            self._name = name

        if not name and ('host' not in self.key.constraints):
            raise ValueError(f'when MachineDependency is called with an InjectionKey, either the key must have a `host` constraint or a name must be given explicitly')

    async def __call__(self, ainjector):
        machine = await ainjector.get_instance_async(self.key)
        if not hasattr(machine, 'start_machine') and hasattr(machine, 'machine'):
            # Allow a dependency to be set on a MachineModel not just a machine
            machine = machine.machine
        await machine.start_machine()
        if self.online is MachineDependency.ONLINE_DEFAULT:
            if machine.machine_running_ssh_online:
                await machine.ssh_online()
        elif self.online:
            await getattr(machine, self.online)()

    @property
    def name(self):
        if hasattr(self, '_name'):
            return self._name
        return self.key.constraints['host']

    def __repr__(self):
        return f'<MachineDependency for {self.name}>'


__all__ += ['MachineDependency']


def disable_system_dependency(injector, dependency):
    "Mask out *dependency* in the scope of *injector*"
    injector.add_provider(dependency.default_instance_injection_key(), None)


__all__ += ['disable_system_dependency']



# Copyright (C) 2018, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage.pytest import *
import os.path
import pytest
from pathlib import Path
from carthage.dependency_injection import AsyncInjector, InjectionKey
from carthage import base_injector, sh
from carthage.config import ConfigLayout
from carthage.image import BtrfsVolume, ReflinkVolume, ContainerVolume, ContainerImage, image_factory
import posix
import gc


@pytest.fixture()
def a_injector():
    if posix.geteuid() != 0:
        pytest.skip("Not running as root; volume tests skipped", )
    injector = base_injector.claim()
    cl = injector(ConfigLayout)
    cl.delete_volumes = True
    yield injector(AsyncInjector)
    gc.collect()


@async_test
async def test_btrfs_volume_base(a_injector, loop):
    cl = await a_injector(ConfigLayout)
    try:
        sh.btrfs("filesystem", "df", cl.image_dir, _bg=False)
    except sh.ErrorReturnCode:
        pytest.skip("image_dir not on btrfs volume")
    assert not os.path.exists(os.path.join(cl.image_dir, "foo")), "The volume directory already exists"
    v = await a_injector(ContainerVolume, implementation=BtrfsVolume, name="foo")
    assert isinstance(v.impl, BtrfsVolume)
    assert str(v.path) == os.path.join(cl.image_dir, v.name)
    assert v.name == "foo"
    v.close()


@async_test
async def test_btrfs_clone(a_injector, loop):
    cl = await a_injector(ConfigLayout)
    try:
        sh.btrfs("filesystem", "df", cl.image_dir, _bg=False)
    except sh.ErrorReturnCode:
        pytest.skip("image_dir is not btrfs")
    vol = await a_injector(ContainerVolume, implementation=BtrfsVolume, name="foo2")
    assert isinstance(vol.impl, BtrfsVolume)
    with open(os.path.join(vol.path, "bar.txt"), "w") as f:
        f.write("This is a file\n")
    v2 = await a_injector(ContainerVolume, implementation=BtrfsVolume, name="clone_foo2", clone_from=vol)
    assert os.path.exists(os.path.join(v2.path, "bar.txt"))


@async_test
async def test_reflink_clone(a_injector):
    cl = await a_injector(ConfigLayout)
    path = Path(cl.image_dir)
    assert not (path / "reflink1").exists()
    rl1 = await a_injector(ContainerVolume, implementation=ReflinkVolume,
                           name=path / "reflink1")
    with open(rl1.path / "bar.txt", "wt") as f:
        f.write("This is a test")
    rl2 = await a_injector(ContainerVolume, implementation=ReflinkVolume,
                           name=path / "reflink2", clone_from=rl1)
    assert rl2.path.joinpath("bar.txt").exists()


@async_test
async def test_container_unpack(a_injector, loop):
    try:
        iv = None
        cl = await a_injector(ConfigLayout)
        cl.base_container_image = Path(__file__).parent.joinpath("resources/base_test.tar.gz")
        iv = await a_injector(ContainerImage, "base")
        path = iv.path
        assert os.path.exists(os.path.join(path, "bin/bash"))
    finally:
        del iv


@pytest.mark.no_rootless
@async_test
async def test_image_unpack(loop, a_injector, vm_image):
    print(vm_image.path)
    with vm_image.image_mounted() as mount:
        assert os.path.exists(os.path.join(mount.rootdir, "bin/bash"))



# Copyright (C) 2018, 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import pytest
from carthage.image import ImageVolume, SshAuthorizedKeyCustomizations
from carthage.vm import InstallQemuAgent
from carthage import base_injector, ConfigLayout, ssh, shutdown_injector
from carthage.dependency_injection import AsyncInjector, DependencyProvider, InjectionKey
from carthage.debian import DebianContainerImage, debian_container_to_vm
from carthage.image import ContainerImage
from carthage.network import Network
from carthage.machine import ssh_origin
from carthage.container import Container, container_image
import os.path
import pytest
import posix
from carthage.pytest import *

pytest_plugins = ('carthage.pytest_plugin',)


@pytest.mark.no_rootless
@pytest.fixture(scope='session')
def test_ainjector(loop):
    if posix.geteuid() != 0:
        pytest.skip("Not running as root; volume tests skipped", )
    try:
        ainjector = base_injector.claim()(AsyncInjector)
        config = loop.run_until_complete(ainjector(ConfigLayout))
        config.delete_volumes = True
        config.persist_local_networking = False
        vol = loop.run_until_complete(ainjector(DebianContainerImage, name="base-debian"))
        vol.config_layout = vol.injector(ConfigLayout)
        vol.config_layout.delete_volumes = False
        loop.run_until_complete(vol.apply_customization(SshAuthorizedKeyCustomizations))
        ainjector.replace_provider(container_image, vol)
        ainjector.replace_provider(loop.run_until_complete(ainjector(Network, 'brint', vlan_id=0)))
        ainjector.replace_provider(ssh_origin, DependencyProvider(None))
        yield ainjector
    finally:
        loop.run_until_complete(shutdown_injector(ainjector))


@pytest.fixture(scope='session')
def vm_image(loop, test_ainjector):
    ainjector = test_ainjector
    debian_container = loop.run_until_complete(ainjector.get_instance_async(container_image))
    config = loop.run_until_complete(ainjector(ConfigLayout))
    image = loop.run_until_complete(ainjector(debian_container_to_vm,
                                              classes="+CLOUD_INIT,SERIAL,OPENROOT",
                                              volume=debian_container,
                                              size="4G",
                                              output=config.vm_image_dir + '/cloud-init.raw'))

    loop.run_until_complete(ainjector.get_instance_async(ssh.SshKey))
    loop.run_until_complete(image.apply_customization(InstallQemuAgent))
    image.config_layout = image.injector(ConfigLayout)
    image.config_layout.delete_volumes = False
    yield image
    image.close()



# Copyright (C) 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio


class Trigger:

    '''An awaitable that can be externally triggered and that by default asserts if not triggered when used as a context manager
'''

    def __init__(self):
        self.future = asyncio.get_event_loop().create_future()
        self.awaited = False

    def __await__(self):
        yield from self.future
        self.awaited = True

    def assert_triggered(self):
        assert self.future.done()
        self.future.result()
        self.awaited = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            if not self.future.done():
                self.future.set_exception(exc_val)
            return False
        assert self.awaited, "This trigger was never awaited"
        return False

    def trigger(self, result=True):
        self.future.set_result(True)



# Copyright (C) 2019, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import collections.abc
import weakref
from ..dependency_injection import *
from . import inventory
from pyVmomi import vim


class ConfigSpecMeta(type):

    def __repr__(self):
        stage = getattr(self, 'stage_for', None)
        if stage is not None:
            stage = getattr(stage, '__name__', stage)
        else:
            stage = "<abstract>"

        return f'<{self.__name__} config stage order={self.order} for {stage}>'


class ConfigSpecStage(metaclass=ConfigSpecMeta):

    '''
    Represents a Stage in building a Vmware configuration.  A number of factors influence which items need to be set in a configuration:

    * For creating an object all items must be set

    * When adding a disk or network to a VM, typically only those items are needed.

    * When cloning a VM, we want to remove all network adapters.

    Beyond that it may be necessary to collect information from parts of the config even if those are not being updated.  As an example, to add a disk, you need the SCSI controllerkey.  So even when  a config item is not edited, it may need to be examined.

    This class represents code that will be run against the config.  The *order* class variable determines which order will be used.

    .. _bag_keys:
    Bag Keys
    ________

    mode
        What sort of configuration mode is being used.  The following modes are defined:

        create
            The initial configuration is being created

        clone
            A VM is being cloned

        clone_disk
            A configspec is being constructed for resizing and adjusting cloned disks.  Only the devices entry from this ConfigSpec will actually be used.

    scsi_key
        SCSI controller key



    '''

    def __init_subclass__(cls, stage_for, order=100, mode='create'):
        '''
        :param stage_for: Which :class:`~inventory.VmwareSpecifiedObject` is this a configuration stage for?

        :param order: When should this stage be run?  Lower numbers are run first.
        :param mode: Either a single mode or a tuple of modes that this config spec works for, or True if for all modes.

'''
        cls.order = order
        if stage_for is None:
            return
        cls.mode = mode
        if isinstance(stage_for, collections.abc.Sequence):
            cls.stage_for = [weakref.proxy(sf) for sf in stage_for]
        else:
            cls.stage_for = [weakref.proxy(stage_for)]
        for sf in cls.stage_for:
            assert issubclass(sf, inventory.VmwareSpecifiedObject)
            if 'config_stages' not in sf.__dict__:
                setattr(sf, 'config_stages', sf.config_stages.copy())
            sf.config_stages.append(cls)
            sf.config_stages.sort(key=lambda c: c.order)

    def __init__(self, obj, bag):
        self.obj = obj
        self.mob = obj.mob
        #: A namespace containing keys shared between stages so that sthages can communicate.
        #:
        #: ..seealso:: `bag_keys`
        self.bag = bag
        if self.mob:
            self.oconfig = self.mob.config
        else:
            self.oconfig = None

    def apply_config(self, config):
        pass

    def __repr__(self):
        return f'<{self.__class__.__name__} config stage for {self.obj} order={self.order}>'


class DeviceSpecStage(ConfigSpecStage, stage_for=None):

    def __init_subclass__(cls, stage_for, dev_classes,
                          **kwargs):
        kwargs.setdefault('mode', ('create', 'reconfig', 'clone'))
        if not isinstance(dev_classes, (tuple, list)):
            dev_classes = (dev_classes,)
        cls.dev_classes = dev_classes
        super().__init_subclass__(stage_for=stage_for, **kwargs)

    @inject(ainjector=AsyncInjector)
    async def apply_config(self, config, *, ainjector):
        if self.oconfig:
            for d in self.oconfig.hardware.device:
                if not isinstance(d, self.dev_classes):
                    continue
                res = await ainjector(self.filter_device, d)
                if res is True:
                    continue
                if isinstance(res, vim.vm.device.VirtualDevice):
                    spec = vim.vm.device.VirtualDeviceSpec()
                    spec.fileOperation = getattr(self, 'file_operation', None)
                    spec.device = res
                    spec.operation = 'edit'
                    config.deviceChange.append(spec)
                elif res is False:
                    spec = vim.vm.device.VirtualDeviceSpec()
                    spec.device = d
                    spec.operation = 'remove'
                    config.deviceChange.append(spec)
                else:
                    raise ValueError(f'Don\'t know how to handle filter_device of {res}')
        # Handle new devices
        for d in await ainjector(self.new_devices, config):
            spec = vim.vm.device.VirtualDeviceSpec()
            spec.device = d
            spec.operation = 'add'
            spec.fileOperation = getattr(self, 'file_operation', None)
            config.deviceChange.append(spec)

    def filter_device(self, d):
        '''
Return *True* to keep the device.

        Return *False* to drop the device from the configuration

        Return A device to change the device

        '''
        return True

    def new_devices(self, config):
        return []


__all__ = ('ConfigSpecStage', 'DeviceSpecStage')



# Copyright (C) 2018, 2019, 2020, 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import contextlib
import contextvars
import enum
import inspect
import typing
import weakref
import collections.abc
import asyncio
import functools
import logging
import traceback
import types
import sys
from dataclasses import dataclass
from .. import tb_utils, event
from .introspection import *
from ..utils import NotPresent

_chatty_modules = {asyncio.futures, asyncio.tasks, sys.modules[__name__]}
logger = logging.getLogger('carthage.dependency_injection')
logger.setLevel('INFO')

#: If true, tracebacks are filtered for better user error messages, but hinding internal state
filter_tracebacks = True


class ReadyState(enum.Enum):
    NOT_READY = 0
    RESOLVED = 1
    READY_PENDING = 2
    READY = 3


instantiate_to_ready = contextvars.ContextVar('instantiate_to_ready', default=True)

log_injection_failed = contextvars.ContextVar('log_injection_failed', default=True)

# While this is needed by InjectableModelType's add_provider, it is
# not part of the public api


def default_injection_key(p):
    if isinstance(p, DependencyProvider):
        raise NotImplementedError
    if (isinstance(p, type) and issubclass(p, Injectable)):
        k = p.default_class_injection_key()
    elif isinstance(p, Injectable):
        k = p.default_instance_injection_key()
    elif isinstance(p, type):
        k = InjectionKey(p)
    else:
        # not a type and not an Injectable
        k = InjectionKey(p.__class__)
    return k


class Injectable:

    '''Represents a class that has dependencies injected into it. By default, the :meth:`__init__` will:

    * Store any keyword arguments corresponding to injection dependencies into instance variables of the same name as the keyword argument

    * Remove these keyword arguments prior to calling the superclass init.

    So for example::

        @inject(router = SiteRouter)
        class Receiver(Injectable): pass

    When ``Receiver`` is instantiated, its instances will have the *router* attribute set.

    It is **recommended** but not required that classes with injected
    dependencies inherit from *Injectable*.  The
    :meth:`satisfies_injection_key` and
    :meth:`supplementary_injection_keys` protocols are only available
    to classes that do inherit from *Injectable*.

    Subclasses that may be mixins and that wish injected dependency handling different than the keyword assignment provided by *Injectable* must inherit from *Injectable*.

    This class does not have :class:`Injector` as an injected dependency.  It is possible to have injected dependencies without doing so.  However, in a dependency is *Injector*, then that injector will be :meth:`claimed <Injector.claim>`.

    '''
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Call inject to deal better with multiple inheritance and dependencies
        try: inject()(cls)
        except NameError:
            if 'inject' not in globals():
                #inject not yet defined
                return
            raise

    def __init__(self, *args, **kwargs):
        autokwargs = set(getattr(self, '_injection_autokwargs', set()))
        for k, d in getattr(self, '_injection_dependencies', {}).items():
            if k in kwargs:
                if d is _injector_injection_key:
                    injector = kwargs.pop(k)
                    setattr(self, k, injector.claim(self))
                else:
                    setattr(self, k, kwargs.pop(k))
                try:
                    autokwargs.remove(k)
                except KeyError:
                    pass

        if autokwargs:
            raise TypeError(f'The following dependencies were not specified: {autokwargs}')

        try:
            super().__init__(*args, **kwargs)
        except TypeError as t:
            if 'object.__init__()' in str(t):
                raise TypeError(f'The following extra arguments were specified: {list(kwargs.keys())}')
            raise

    def close(self, canceled_futures=None):
        if hasattr(self, 'injector'):
            self.injector.close(canceled_futures)

    @classmethod
    def supplementary_injection_keys(cls, k: InjectionKey):
        '''
        Returns  an iteration of :class:`InjectionKeys <InjectionKey>` that should be added  to an injector when this class is added.  The current injection key is taken as an argument so that constraints applied can modify what keys are added.


        '''
        for c in cls.__mro__:
            if c in (Injectable, AsyncInjectable):
                continue
            if issubclass(c, Injectable) and c != k.target:
                yield InjectionKey(c)
                if k.constraints:
                    yield InjectionKey(c, **k.constraints)
            elif c is k.target and k.constraints:
                yield InjectionKey(c)

    @classmethod
    def satisfies_injection_key(cls, k):
        if k == InjectionKey(cls):
            return True
        if isinstance(k.target, (str, tuple)):
            return True
        return issubclass(cls, k.target)

    @classmethod
    def default_class_injection_key(cls):
        '''Called by :meth:`Injector.add_provider()` in the single argument form to get the injection key to use when this class is added to provide a dependency.'''
        return InjectionKey(cls)

    def default_instance_injection_key(self):
        "Called when an instance of an injectable is used to add a dependency provider bby the single argument form of :meth:`Injectable.add_provider()`"
        return InjectionKey(self.__class__)

class DependencyProvider:
    __slots__ = ('provider',
                 'allow_multiple',
                 'close',
                 'instantiation_contexts',
                 'keys',
                 '_creation_tb',
                 )

    def __init__(self, provider, allow_multiple=False, close=True):
        self.provider = provider
        self.allow_multiple = allow_multiple
        self.close = close
        self.keys = set()
        self.instantiation_contexts = set()
        self._creation_tb = traceback.extract_stack()[:-1]

    def __repr__(self):
        return "<DependencyProvider allow_multiple={}: {}>".format(
            self.allow_multiple, repr(self.provider))

    @property
    def is_factory(self):
        return (isinstance(self.provider, type) and issubclass(self.provider, Injectable)) \
            or asyncio.iscoroutinefunction(self.provider) \
            or directly_has_dependencies(self.provider)

    @property
    def needs_quote(self):
        return self.is_factory \
            or asyncio.isfuture(self.provider) \


    def record_instantiation(self, instance, k, satisfy_against, final):
        dp = satisfy_against._providers.setdefault(
            k, DependencyProvider(
                instance, self.allow_multiple, close=self.close))
        assert dp.needs_quote or dp.provider is instance
        dp.provider = instance
        if final:
            # If requested again, would try to instantiate again.
            if dp.needs_quote:
                dp.provider = dependency_quote(dp.provider)
        return dp


class InjectionFailed(RuntimeError):

    def __init__(self, context):
        super().__init__(f"Error {str(context)}")
        ctx = context
        dependency_path = []
        while ctx:
            try:
                dependency_path.insert(0,ctx.key)
                ctx = ctx.parent
            except AttributeError:
                ctx = ctx.parent
        try:
            self.failed_dependency = dependency_path[-1]
        except IndexError: pass
        self.dependency_path = tuple(dependency_path)
        

class ExistingProvider(RuntimeError):

    def __init__(self, k, old_p, new_p):
        self.existing_key = k
        self.old_provider = old_p
        self.new_provider = new_p
        s = f'Unable to add provider {new_p.provider} for {k}: already registered to {old_p.provider}.\n\n'
        s = s + 'The previous provider was created here:\n\n'
        s = s + '> ' + '> '.join(traceback.format_list(old_p._creation_tb))
        super().__init__(s)


class InjectorClosed(RuntimeError):
    pass


class AsyncRequired(RuntimeError):

    def __init__(self, msg, context):
        super().__init__(f'{msg} {context.description if context else ""}')


# Note that after @inject is defined, this class is redecorated to take parent_injector as a dependency so that
#    injector = sub_injector(Injector)
# works
class Injector(Injectable, event.EventListener):

    def __init__(self, *providers,
                 parent_injector=None):
        self._providers = {}
        self._pending = weakref.WeakSet()
        self.closed = False
        self._closing = False
        if parent_injector is None and len(providers) > 0:
            if isinstance(providers[0], Injector):
                parent_injector = providers[0]
                providers = providers[1:]

        self.parent_injector = parent_injector
        self.claimed_by = None
        if self.parent_injector:
            event_scope = self.parent_injector._event_scope
            event_scope.add_child(parent_injector, self)
        else:
            event_scope = None
        super().__init__(event_scope=event_scope)
        for p in providers:
            self.add_provider(p)
        self.add_provider(self)  # Make sure we can inject an Injector
        self.add_provider(InjectionKey(AsyncInjector), AsyncInjector, allow_multiple=True)

    def claim(self, claimed_by=True):
        '''
        Take ownership of the injector.

        :param claimed_by: Either *True* or an object that this injector is marked as belonging to.

        Returns either *self* or a new subinjector.

        '''
        if self.claimed_by:
            return self(type(self)).claim(claimed_by)
        else:
            if claimed_by is True or isinstance(claimed_by, str):
                self.claimed_by = claimed_by
            else:
                self.claimed_by = weakref.ref(claimed_by)
            return self

    def add_provider(self, k, p=None, *,
                     allow_multiple=False,
                     close=True,
                     replace=False):
        '''Add a provider for a dependency

        Either called as ``add_provider(provider)`` or
        ``add_provider(injection_key, provider)``\\ .  In the first form, a key is
        automatically constructed.  For :class:`Injectable` objects, see :meth:`Injectable.default_instance_injection_key()`.  For *Injectable* types, see :meth:`Injectable.default_class_injection_key()`.  For other objects, the unconstrained injection key for types (or for the type of the object) is used.

        :param allow_multiple: If true, then this provider may be instantiated multiple times in sub-injectors.  If false (the default) then the provider will be instantiated on the injector where it is added and used by all sub-injectors.

        :param close: If true (the default), then closing the injector will close or cancel this provider.  If false, then the provider will not be deallocated.  As an example, if the :class:`asyncio.AbstractEventLoop` is added as a provider, but closing this injector should not close the loop and end all async operations, then close can be set to false.

        :param replace: If True, an existing provider is being updated.  :meth:`replace_provider` is a convenience function for calling :meth:`add_provider` with *replace* set to True.  Replacing providers may lead to inconsistent results if the provider has already been injected to fill a dependency in a constructed object.

'''
        if p is None and not isinstance(k, InjectionKey):
            p, k = k, p  # swap; we construct the key later

        if k is None:
            k = default_injection_key(p)
        if not isinstance(p, DependencyProvider):
            p = DependencyProvider(p, allow_multiple=allow_multiple, close=close)
        assert isinstance(k, InjectionKey)
        if k in self:
            if p is self._get(k):
                return k
            existing_provider = self._get(k)
            if replace:
                existing_provider.provider = p.provider
                existing_provider.keys.add(k)
            else:
                raise ExistingProvider(k, existing_provider, p)
        else:
            self._providers[k] = p
            p.keys.add(k)
        for k2 in k.supplementary_injection_keys(p.provider):
            if k2 not in self:
                self._providers[k2] = p
                p.keys.add(k2)
        self.emit_event(
            k, "add_provider",
            p.provider,
            replace=replace, close=close,
            allow_multiple=allow_multiple,
            inspector=InjectedDependencyInspector(injector=self, key=k, provider=p),
            other_keys=p.keys,
            adl_keys=p.keys | {InjectionKey(Injector)})
        return k

    def replace_provider(self, *args, **kwargs):
        return self.add_provider(*args, **kwargs, replace=True)

    def _get(self, k):
        return self._providers[k]

    def _get_parent(self, k):
        # Returns  DependencyProvider, instantiation_target
        injector = self
        while injector is not None:
            try:
                # If the key allows multiple providers, then
                # satisfy against ourself and store the result in
                # ourself.  Otherwise if a single provider is
                # required, then satisfy against the injector
                # where the key is introduced and store there.
                p = injector._providers[k]
                return p, (self if p.allow_multiple else injector)
            except KeyError:
                injector = injector.parent_injector
        raise KeyError("{} not found".format(k))

    def injector_containing(self, k):
        '''
Return the first injector in our parent chain containing *k* or None if there is no such injector.

        If *k* has not yet been instantiated, this injector would be the one against which the instantiation is recorded unless the provider was added with the *allow_multiple* argument to :meth:`add_provider()`.
        '''
        if not isinstance(k, InjectionKey):
            k = InjectionKey(k)
        injector = self
        while injector and not k in injector:
            injector = injector.parent_injector
        if injector and (k in injector):
            return injector
        return None

    def __contains__(self, k):
        if not isinstance(k, InjectionKey):
            k = InjectionKey(k)
        return k in self._providers

    def _check_closed(self):
        if self.closed:
            raise InjectorClosed("Injector is closed")

    def filter(self,
               target: type,
               predicate: typing.Union[list, typing.Callable] = None,
               stop_at: Injector = None):
        '''
        :return: list of :class:`InjectionKey` with target type of
            *target* and satisfying *predicate* in the current injector
            and its parents.

        :param predicate: A list of constraints that must all be present in the key, or a callable that returns true if the key should be included.

        :param target: A target type to filter against.

        :param stop_at: An injector which must be a parent of this injector.  Do not progress past that injector in finding keys.  So if *stop_at* is *self*, only locally registered keys are returned.


        Example usage would be to find all registered plugins similar to the following::

            plugin_keys =injector.filter(CarthagePlugin, ['name'])

        '''
        def filter_for_constraints(k):
            return all(map(lambda c: c in k.constraints, constraints))
        if isinstance(predicate, list):
            constraints = predicate
            predicate = filter_for_constraints
        if stop_at == self:
            result = {}  # stop here
        elif stop_at and not self.parent_injector:
            raise ValueError(f'{stop_at} was not in the parent chain')
        elif self.parent_injector:
            result = {k: True for k in self.parent_injector.filter(target, predicate, stop_at=stop_at)}
        else:  # no stop_at; ended chain
            result = {}
        result.update({k: True for k in self._providers.keys() if ((not target) or   k.target is target ) and predicate(k)})
        return list(result.keys())

    def filter_instantiate(self, target, predicate, *, stop_at=None, ready=False):
        '''
        Like :meth:`filter` but an iterator returning tuples of keys instance.
'''
        for k in self.filter(target, predicate, stop_at=stop_at):
            if ready is not None:
                k = InjectionKey(k, _ready=ready)
            res = self.get_instance(k)
            if res is not None:
                yield k, res

    def inspect(self, *,key_filter=None, include_parent:bool = False):
        '''
        Inspect the contents of this injector.
        This is a generator yielding key, :class:`~InjectedDependencyInspector` pairs for each dependency provided by the injector.

        :param key_filter: A function that takes a key and returns true if  the inspector should inspect the dependency for this key.

        :param include_parent: By default only this injector is  covered; if True, then the inspection recurses into the parent.

        '''
        if key_filter:
            already_included = set(filter(key_filter, self._providers.keys()))
            parent_filter = lambda k: k not in already_included and key_filter(k)
        else:
            already_included = set(self._providers.keys())
            parent_filter = lambda k: k not in already_included
        yield from ((k, InjectedDependencyInspector(
            key=k, injector=self, provider=self._providers[k]
            )) for k in already_included)
        if include_parent and self.parent_injector: 
            yield from self.parent_injector.inspect(
                key_filter=parent_filter, include_parent=True)

    @property
    def child_injectors(self):
        return frozenset(self._event_scope.children[id(self)])
    
            

            
    def __call__(self, cls, *args, **kwargs):
        '''Construct an instance of cls using the providers in this injector.
        Instantiate providers as needed.  In general a sub-injector is not
        constructed.  However if any keyword arguments pased in specify a
        dependency, then construct an injector for that.  Keyword arguments
        and arguments are passed to the class to construct the object.  If
        keyword arguments do specify a dep.dependency, they must satisfy the
        InjectionKey involved.
'''
        self._check_closed()
        return self._instantiate(
            cls, *args, **kwargs,
            _loop=None,
            _placement=None,
            _interim_placement=None)

    def get_instance(self, k,
                     placement=None,
                     loop=None, futures=None):
        '''
        Get an instance satisfying a given :class:`InjectionKey`.

        :param loop: An asyncio loop.  If provided, then asynchronous activities  can take place.
        :param placement: A function taking one argument.  Once the dependency is resolved, this function will be called with the result.  More convenient for asyncronous  operations.
        :param futures: If the result cannot be determined immediately, then a future will be added to this list.

        Note that If any of *loop* or *futures*,  are provided, both must be provided.  If *loop* is provided, then the return may be a future.

        '''
        if loop:
            assert futures is not None

        def do_place(res):
            provider.record_instantiation(res, k, satisfy_against, final=True)
            instantiation_context.final()
            if placement:
                placement(res)

        def do_interim_place(res):
            provider.record_instantiation(res, k, satisfy_against, final=False)
            instantiation_context.progress()

        logger.debug("Looking up provider for {}".format(k))

        if not isinstance(k, InjectionKey):
            k = InjectionKey(k)
        try:
            provider, satisfy_against = self._get_parent(k)
        except KeyError:
            self._check_closed()
            if k.optional:
                res = None if k.optional is True else k.optional
                if placement:
                    placement(res)
                return res
            raise KeyError("No dependency for {}".format(k)) from None
        mark_instantiation_done = True
        with InstantiationContext(
                satisfy_against, self, k, provider,
                k.ready if (k.ready is not None) else instantiate_to_ready.get()) as instantiation_context:
            try:
                if k.ready is not None:
                    ready_reset = instantiate_to_ready.set(k.ready)
                else:
                    ready_reset = None
                to_ready = instantiate_to_ready.get()
                result = provider.provider
                if isinstance(result, dependency_quote):
                    if placement:
                        placement(result.value)
                    return result.value
                elif isinstance(result, asyncio.Future) \
                        or (not provider.is_factory and to_ready and not is_obj_ready(result)):
                    # We need to handle the async object through
                    # _handle_async; if it is a future it may resolve
                    # to an AsyncInjectable, and if it needs to become
                    # ready we need to allow for that.  We also need
                    # to set up placement.
                    if loop is None:
                        raise AsyncRequired(
                            f'{k} requires asynchronous instantiation',
                            current_instantiation())
                    result = loop.create_task(self._handle_async(
                        result,
                        placement=do_place, loop=loop,
                        mark_instantiation_done=True, instantiation_context=instantiation_context))
                    mark_instantiation_done = False
                    result.set_name('handle ' + str(current_instantiation()))
                    futures.append(result)
                    instantiation_context.progress()
                    return result
                elif provider.is_factory:
                    result = satisfy_against._instantiate(
                        result,
                        _loop=loop,
                        _placement=do_place,
                        _interim_placement=do_interim_place,
                        _instantiation_context=instantiation_context,
                    )
                if isinstance(result, asyncio.Future):
                    if loop is None:
                        raise AsyncRequired(
                            f'{k} requires asynchronous instantiation',
                            current_instantiation())
                    futures.append(result)
                    provider.record_instantiation(result, k, satisfy_against, final=False)
                    instantiation_context.progress()
                    mark_instantiation_done = False
                    return result

            finally:
                if ready_reset is not None:
                    instantiate_to_ready.reset(ready_reset)
                if mark_instantiation_done:
                    instantiation_context.done()

            # Either not a future or not a factory
            if placement:
                placement(result)
            return result

    def _instantiate(self, cls, *args,
                     _loop,
                     _placement,
                     _interim_placement,
                     _instantiation_context=None,
                     **kwargs):
        # _loop if  present means we can return something for which _is_async will return True
        # There are complicated interactions between the
        # DependencyProvider machinery and the InstantiationContext
        # machinery.  An instantiationcontext is marked done when  the
        # instantiation has completed or failed and no other events
        # will be logged against this context.  A DpendencyProvider is
        # recorded as final when the underlying object providing the
        # dependency will no longer change.  That doesn't mean the
        # instantiation is done though.  The most common case where a
        # DependencyProvider can be final but where the instantiation
        # is not done is where async_become_ready needs to be
        # called.
        self._check_closed()

        def handle_result(mark_instantiation_done):
            # Called when all kwargs are populated
            try:
                try:
                    res = cls(*args, **kwargs)
                except TypeError as e:
                    raise TypeError(f'Error constructing {cls}:') from e
                if self._is_async(res):
                    if not _loop:
                        if isinstance(res, collections.abc.Coroutine):
                            res.close()
                        raise AsyncRequired(
                            "Asynchronous dependency injected into non-asynchronous context",
                            current_instantiation())

                    future = asyncio.ensure_future(self._handle_async(
                        res,
                        placement=_placement,
                        loop=_loop, mark_instantiation_done=mark_instantiation_done,
                        instantiation_context=_instantiation_context))
                    if _interim_placement:
                        _interim_placement(future)
                    self._pending.add(future)

                    if _instantiation_context:
                        future.set_name(f'async {_instantiation_context.description}')
                    else:
                        future.set_name(f'Handle async  {res} for {self}')
                    return future
                else:
                    if _placement:
                        _placement(res)
                    return res
            except AsyncRequired:
                raise
            except Exception as e:
                if filter_tracebacks:
                    tb_utils.filter_chatty_modules(e, _chatty_modules, None)
                if _instantiation_context:
                    failed_instantiation(_instantiation_context, e)
                if _instantiation_context and not isinstance(e, InjectionFailed):
                    tb_utils.filter_before_here(e)
                    if log_injection_failed.get():
                        logger.exception(f'Error resolving dependency for {current_instantiation()}')
                    raise InjectionFailed(current_instantiation()) from e
                else:
                    raise

        async def callback(futures):
            try:
                await asyncio.gather(*futures)
                # since that succeeded, allkwargs are placed
                res = handle_result(mark_instantiation_done=False)
                if isinstance(res, asyncio.Future):
                    return await res
                return res
            finally:
                if _instantiation_context:
                    _instantiation_context.done()

        def kwarg_place(k):
            def collect(res):
                if res is not NotPresent:
                    kwargs[k] = res
            return collect
        try:
            futures = []

            dks = set(filter(
                lambda k: cls._injection_dependencies[k] is not None,
                cls._injection_dependencies.keys()))
        except AttributeError:
            dks = set()
        injector = self  # or sub_injector if created
        sub_injector = None
        kwarg_dependencies = set(kwargs.keys()) & dks
        try:  # clean up sub_injector
            if kwarg_dependencies:
                sub_injector = (type(self))(self)
                injector = sub_injector
                for k in kwarg_dependencies:
                    provider = kwargs.pop(k)
                    dependency = cls._injection_dependencies[k]
                    if isinstance(provider, Injectable) and not provider.satisfies_injection_key(dependency):
                        raise UnsatisfactoryDependency(dependency, provider)
                    sub_injector.add_provider(dependency, provider, close=False)
            for k, d in (cls._injection_dependencies.items()) if dks else []:
                if d is None:
                    continue
                injector.get_instance(d, placement=kwarg_place(k),
                                      loop=_loop, futures=futures)
            if futures:
                fut = asyncio.ensure_future(callback(futures))
                if current_instantiation():
                    fut.set_name(f'{current_instantiation()}')
                else:
                    fut.set_name(f'Instantiate {cls} for {self}')
                return fut

            else:
                res = handle_result(mark_instantiation_done=(_instantiation_context is not None))
                return res
        finally:
            # Perhaps some day we need to clean up something about the sub_injector
            pass

    def _is_async(self, p):
        if isinstance(p, (collections.abc.Coroutine,
                          asyncio.Future)):
            return True
        elif isinstance(p, AsyncInjectable) and p._async_ready_state != ReadyState.READY:
            to_ready = instantiate_to_ready.get()
            return (p._async_ready_state == ReadyState.NOT_READY) or to_ready
        return False

    async def _handle_async(self, p,
                            placement,
                            loop,
                            instantiation_context=None,
                            mark_instantiation_done=True):
        if not hasattr(self, 'loop'):
            self.loop = loop
        try:
            if isinstance(p, collections.abc.Coroutine):
                res = await p
            elif isinstance(p, AsyncInjectable):
                res = await self._handle_async_injectable(p, placement=placement,
                                                          mark_instantiation_done=False, instantiation_context=instantiation_context)
            elif isinstance(p, asyncio.Future):
                # We may need to re-add an instantiation to
                # waiting dependencies in the instantiations that
                # are moving to ready.
                if mark_instantiation_done:
                    current_instantiation().progress()
                res = await p
                if isinstance(res, AsyncInjectable) and res._async_ready_state != ReadyState.READY:
                    res = await self._handle_async_injectable(
                        res, placement=placement,
                        mark_instantiation_done=False, instantiation_context=instantiation_context)
            else:
                raise RuntimeError('_is_async returned True when _handle_async cannot handle')
            if placement:
                placement(res)
            return res
        except asyncio.CancelledError:
            if hasattr(p, 'injector'):
                await shutdown_injector(p.injector)
            raise
        except Exception as e:
            if filter_tracebacks:
                tb_utils.filter_chatty_modules(e, _chatty_modules, None)
            if instantiation_context:
                failed_instantiation(instantiation_context, e)
            if instantiation_context and not isinstance(e, InjectionFailed):
                tb_utils.filter_before_here(e)
                if log_injection_failed.get():
                    logger.exception(f'Error resolving dependency for {current_instantiation()}')
                raise InjectionFailed(instantiation_context) from e
            else:
                raise

        finally:
            if mark_instantiation_done and current_instantiation():
                current_instantiation().done()

    async def _handle_async_injectable(self, obj, placement, resolv=True,
                                       instantiation_context=None, mark_instantiation_done=False):
        try:
            # Don't bother running the resolve protocol for the base case
            if resolv and (obj._async_ready_state == ReadyState.NOT_READY):
                res = await obj.async_resolve()
                if res is None or res is obj:
                    if obj._async_ready_state == ReadyState.NOT_READY:
                        obj._async_ready_state = ReadyState.RESOLVED
                    res = obj
                if self._is_async(res):
                    return await self._handle_async(
                        res, placement=placement,
                        loop=self.loop,
                        mark_instantiation_done=False, instantiation_context=instantiation_context)
                else:
                    return res
            else:  # no resolution required
                if instantiate_to_ready.get():
                    if placement:
                        placement(obj)
                    await obj.async_become_ready()
                    if not obj._async_ready_state == ReadyState.READY:
                        raise RuntimeError(
                            f"async_ready for {obj.__class__.__name__} must chain back to AsyncInjectable.async_ready.")

                return obj
        except asyncio.CancelledError:
            if hasattr(obj, 'injector'):
                await shutdown_injector(obj.injector)
            raise
        finally:
            if mark_instantiation_done:
                current_instantiation().done()

    def close(self, canceled_futures=None):
        '''
        Close all subinjectors or providers

        For every provider registered with this injector, call :meth:`close` if it is exists.  Then clear out all providers.  Note that this will also close sub-injectors of providers.

        If using `AsyncInjector`, it is better to call :func:`shutdown_injector` to cancel any running asynchronous tasks.

        If the provider's :meth:`close` method takes an argument called *canceled_futures* then the *canceled_futures* argument will be passed down.
        '''

        for f in self._pending:
            try:
                f.cancel()
            except BaseException:
                pass
        self._pending.clear()
        if self._closing:
            return
        self._closing = True
        providers = list(self._providers.values())
        for p in providers:
            if p.provider is self or not p.close:
                continue
            if hasattr(p.provider, 'close'):
                try:
                    _call_close(p.provider, canceled_futures)
                except Exception:
                    logger.exception("Error closing {}".format(p))
            elif asyncio.isfuture(p.provider):
                p.provider.cancel()
                if canceled_futures is not None:
                    canceled_futures.append(p.provider)
        self.closed = True
        del providers
        self._providers.clear()
        self.parent_injector = None

    def __del__(self):
        if not self.closed:
            self.close()

    def __repr__(self):
        claim_str = ""
        if self.claimed_by is True:
            claim_str = f'claimed id: {id(self)}'
        elif isinstance(self.claimed_by, str):
            claim_str = self.claimed_by
        elif self.claimed_by is None:
            claim_str = f'unclaimed id: {id(self)}'
        elif self.claimed_by() is None:
            claim_str = "claimed by dead object"
        else:
            try:
                claim_str = f'claimed by {repr(self.claimed_by())}'
            except Exception:
                claim_str = f'claimed by object with id {id(self.claimed_by())}'
        closed_str = 'CLOSED ' if self.closed else ''
        return f'<{closed_str}{self.__class__.__name__} {claim_str}>'

    @property
    def is_claimed(self):
        return self.claimed_by is not None


_INJECTION_KEY_DEFAULTS = {
    'optional': False,
    'globally_unique': False,
    'ready': None}


class InjectionKey:

    '''
    Represents information about what is requested to satisfy a dependency.

    :param target: A type or other object representing what is desired.

        * A type indicating an object of that type is desired

        * An object such as a string that is a unique identifier for what is desired

    :param _optional: If *True*, then if no provider for the dependency is registered, None will be passed rather than raising; if some other true value, that value will be returned by default.  If *NotPresent*, then no kwarg will be specified if the dependency is not provided. Because of the historical API, there is no current way to have an optional key default to *True*.  The *optional* parameter is read from the call to :meth:`Injector.add_provider`.

    :param _ready:  If None (the default), then use the same readyness as the object into which this is being injected (or full readyness if this is a base operation).  If True, then to satisfy this dependency, the provided object must be fully ready.  If False, then a not ready object is preferred.

    :param _globally_unique:  If true, then this provided dependency (typically in a call to :func:`carthage.modeling.provides`) is globally unique and need not be modified during :class:`container propagation <carthage.modeling.ModelingContainer>`.

    '''

    POSSIBLE_PARAMETERS = frozenset(
        set(map(
            lambda k: '_' + k, _INJECTION_KEY_DEFAULTS))
        | {'optional'})

    _target_injection_keys = weakref.WeakKeyDictionary()

    def __new__(cls, target_, *, require_type=False, **constraints):
        assert (cls is InjectionKey) or set(constraints) - \
            cls.POSSIBLE_PARAMETERS, "You cannot subclass InjectionKey with empty constraints"
        if require_type and not isinstance(target_, type):
            raise TypeError(
                'Only types can be used as implicit injection keys; if this is intended then construct the injection key explicitly')
        if isinstance(target_, InjectionKey):
            # mostly so you can take an existing injection key and mark it optional
            new_constraints = dict(target_.constraints)
            new_constraints.update(constraints)
            constraints = new_constraints
            target_ = target_.target
        if (not constraints):
            if target_ in cls._target_injection_keys:
                return cls._target_injection_keys[target_]
        self = super().__new__(cls)
        customized = bool(constraints)
        if '_optional' not in constraints:
            try:
                constraints['_optional'] = constraints.pop('optional')
            except KeyError:
                pass
        for k in _INJECTION_KEY_DEFAULTS:
            self.__dict__[k] = constraints.pop(
                '_' + k, _INJECTION_KEY_DEFAULTS[k])

        self.__dict__['constraints'] = dict(constraints)
        self.__dict__['target'] = target_
        if (not customized) and not isinstance(target_, (str, int, float)):
            cls._target_injection_keys[target_] = self
        return self

    def __getattr__(self, k):
        if k in self.__dict__:
            return self.__dict__[k]
        if k in self.__dict__['constraints']:
            return self.__dict__['constraints'][k]
        raise AttributeError

    def __repr__(self):
        r = "InjectionKey({}".format(
            self.target.__name__ if isinstance(self.target, type) else repr(self.target))
        if self.ready is not None:
            r += f', _ready={self.ready}'
        for k, v in self.constraints.items():
            r += f', {str(k)}={repr(v)}'
        return r + ")"

    def __setattr__(self, k, v):
        raise TypeError('InjectionKeys are immutable')

    def __hash__(self):
        return hash(self.target) + sum([hash(k) for k in self.constraints.keys()]) + \
            sum([hash(v) for v in self.constraints.values()])

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if self.target != other.target:
            return False
        if len(self.constraints) != len(other.constraints):
            return False
        if all(map(lambda k: self.constraints[k] == other.constraints[k], self.constraints.keys())):
            return True
        return False

    def supplementary_injection_keys(self, p):
        if (isinstance(p, type) and issubclass(p, Injectable)) or \
           isinstance(p, Injectable):
            yield from p.supplementary_injection_keys(self)
        else:
            if p.__class__ in (int, float, str, list, tuple, types.FunctionType):
                return
            for c in p.__class__.__mro__:
                if c is p.__class__:
                    continue
                yield InjectionKey(c)


# Used in Injector.__init__
_injector_injection_key = InjectionKey(Injector)


@dataclass
class UnsatisfactoryDependency(RuntimeError):
    dependency: InjectionKey
    provider: DependencyProvider
    reason: str = None


def inject(**dependencies):
    '''A dhecorator to indicate that a function requires dependencies:

    Sample Usage::

        @inject(injector = Injector,
            router = InjectionKey(SiteRouter, site ='cambridge'))
        def testfn(injector, router): pass

    Can be applied to classes or functions.  Note that when an injector is used, dependencies will be resolved through the injector  even if they are supplied directly.  For example in::

        @inject(dependency = InjectedClass)
       def func(dependency):
             assert isinstance(dependency, InjectedClass)


        injector(func, dependency=InjectedClass)

    The assertion will be true if *InjectedClass* is an :class:`Injectable` because the injector will instantiate the class.  Resolution provided by the injector includes:

    #. Instantiating subclasses of :class:`Injectable` and providing their dependencies.

    #. Calling :meth:`~AsyncInjectable.async_ready` on :class:`AsyncInjectable`.

    '''
    def convert_to_key(deps):
        for k, v in deps.items():
            if isinstance(v, InjectionKey):
                yield k, v
            elif v is None:
                yield k, v
            else:
                yield k, InjectionKey(v, require_type=True)

    def init_from_bases(c, dependencies, autokwargs):
        for b in c.__bases__:
            if hasattr(b, "_injection_dependencies"):
                autokwargs -= b._injection_this_level
                dependencies.update(b._injection_dependencies)
                autokwargs |= b._injection_autokwargs

    def wrap(fn):
        if getattr(fn, '_injection_error', None):
            #Set _injection_error on decorators or other things that might be mistakenly targets of inject.
            raise TypeError(fn._injection_error)
        
        if (not hasattr(fn, '_injection_dependencies')) or (isinstance(
                fn, type) and '_injection_dependencies' not in fn.__dict__):
            fn._injection_dependencies = dict()
            fn._injection_this_level = set()
            fn._injection_autokwargs = set()
            if isinstance(fn, type):
                init_from_bases(fn, fn._injection_dependencies, fn._injection_autokwargs)

        for k, v in convert_to_key(dependencies):
            try:
                fn._injection_autokwargs.remove(k)
            except KeyError:
                pass
            if v is not None:
                # So autokwargs doesn't include it
                fn._injection_this_level.add(k)
            fn._injection_dependencies[k] = v
        return fn
    return wrap


def inject_autokwargs(**dependencies):
    '''
    Like :func:`inject` but explicitly marks that the keywords are expected to fall through to :meth:`Injectable.__init__`
    Applies to all dependencies at the current level so can be used either like::

        @inject_autokwargs(foo = bar)
        class baz(Injectable):

    or like::

        @inject_autokwargs()
        @inject(foo = bar)
        class baz(Injectable):

    :class:`InjectionKey`s with *_optional* set to NotPresent are not required.
    '''
    def wrap(cls):
        inject(**dependencies)(cls)
        cls._injection_autokwargs |= set(filter(
            lambda k: cls._injection_dependencies[k].optional is not NotPresent, cls._injection_this_level))
        return cls
    return wrap


def copy_and_inject(_wraps=None, **kwargs):
    "Like inject but makes a copy of the dependencies first; typically used when wrapping an injector"
    def wrap(fn):
        if hasattr(fn, '_injection_dependencies'):
            fn._injection_dependencies = fn._injection_dependencies.copy()
        return inject(**kwargs)(fn)
    if _wraps is not None:
        return wrap(_wraps)
    else:
        return wrap


Injector = inject(parent_injector=Injector)(Injector)


def partial_with_dependencies(func, *args, **kwargs):
    '''Partially aply arguments and keep injected dependencies

    Like :class:`functools.partial` except also preserves dependencies.
    Used typically when passing the result of *partial* to
    :meth:`Injector.add_provider`

    This implementation assumes that no dependencies are removed by
    passing arguments into partial that specify one of the injected
    dependencies.

    '''
    p = functools.partial(func, *args, **kwargs)
    try:
        p._injection_dependencies = func._injection_dependencies
    except AttributeError:
        pass
    return p


def directly_has_dependencies(f):
    '''

    :return: True if *f* directly has injection dependencies applied.  Not true for an object of a class even if that class has dependencies.

    '''
    if not hasattr(f, '__dict__'):
        return False
    return '_injection_dependencies' in f.__dict__


class dependency_quote:

    '''
Indicate that a value (or dependency provider) should not be subject to injcoter resolution.  Used like::

        injector(func, override = dependency_quote(self))

    For example if *self* is an object that is not yet :meth:`~AsyncInjectable.async_ready` as part of bootstrapping.  Can be used both in a keyword argument to an injector call or in calls to :meth:`Injector.add_provider`.

'''

    __slots__ = ['value', '__transclusion_key__']

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f'dependency_quote({repr(self.value)})'

   #########################################
   # Asynchronous support:


@inject_autokwargs(
    injector=Injector)
class AsyncInjectable(Injectable):

    '''

    An :class:`Injectable` that supports asyncronous operations as part of making a dependency available.  This happens in several phases:

    * Prior to construction, all the dependencies of the *Injectable* are prepared.

    * :meth:`async_resolve` is called.  This asynchronous method can return a different object, which entirely replaces this object as the provider of the dependency.  The *async_resolve* protocol is intended for cases where figuring out which object will provide a dependency requires asynchronous operations.  In many cases :meth:`async_resolve` returns *self*.

    * Call :meth:`async_ready` to prepare this object.  This may include doing things like running :func:`~carthage.setup_tasks.setup_task`.

'''

    _async_ready_state: ReadyState

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # superclass claims the injector for us.
        if hasattr(self, 'injector'):
            self.ainjector = self.injector(AsyncInjector)
        if self.async_resolve.__func__ is AsyncInjectable.async_resolve:
            if self.async_ready.__func__ is AsyncInjectable.async_ready:
                self._async_ready_state = ReadyState.READY
            else:
                self._async_ready_state = ReadyState.RESOLVED
        else:  # has nontrivial async_resolve
            self._async_ready_state = ReadyState.NOT_READY

    async def async_ready(self):
        self._async_ready_state = ReadyState.READY
        return self

    async def async_resolve(self):
        '''Returns None or an object that should replace *self* in providing dependencies.'''
        return None

    async def async_become_ready(self, cycle_set=None):

        '''
        Interface point to request that a dependency be fully ready (:meth:`async_ready` is called exactly once).  This function manages tracking to make sure that async_ready is called once, and waits for that call if needed.
        This method also makes sure that dependencies registered with :func:`.inject` are made ready before :meth:`async_ready` is called.
        Subclasses should not override this method but instead should override :meth:`async_ready`.
        '''
        with AsyncBecomeReadyContext(self):
            if self._async_ready_state == ReadyState.NOT_READY:
                raise RuntimeError("Resolution should have already happened")
            elif self._async_ready_state == ReadyState.RESOLVED:
                self._ready_future = asyncio.ensure_future(_handle_async_deps(self, cycle_set))
                self._ready_future.set_name(f'Async dependencies for {self}')
                self._async_ready_state = ReadyState.READY_PENDING
                try:
                    return await self._ready_future
                except BaseException:
                    self._async_ready_state = ReadyState.RESOLVED
                    raise
                finally:
                    del self._ready_future
            elif self._async_ready_state == ReadyState.READY_PENDING:
                return await asyncio.shield(self._ready_future)
            else:
                return


@inject(loop=asyncio.AbstractEventLoop, injector=Injector)
class AsyncInjector(Injectable):

    '''An asynchronous injector.  AsyncInjector is not a subclass of
    Injector because AsyncInjector's call function is a coroutine and
    so it has an incompatible interface.  In other ways the classes
    should behave the same.

    This class overrides :class:`Injectable`'s behavior of claiming the injector.  Instead, if you construct an *AsyncInjector* you get exactly what you asked for: an *AsyncInjector* that maps directly onto the injector you construct.  Note however that when an *AsyncInjector* is constructed by :class:`AsyncInjectable`, the injector is claimed properly.

'''

    def __init__(self, injector, loop):
        self.injector = injector
        self.injector.replace_provider(self)
        self.loop = loop
        # For methods that injector has but we do not, then call the method on our
        # injector.  This is a lot like inheritance but does not make us a
        # subclass.
        for k in list(Injector.__dict__.keys()) + list(event.EventListener.__dict__.keys()):
            if not isinstance(getattr(Injector, k), types.FunctionType):
                continue

            if hasattr(self, k):
                continue
            setattr(self, k, getattr(self.injector, k))

    def claim(self, claimed_by=True):
        if self.injector.is_claimed:
            return type(self)(injector=self.injector.claim(claimed_by),
                              loop=self.loop)
        else:
            assert self.injector.claim(claimed_by) is self.injector
            return self

    def __repr__(self):
        return f'<Async Injector Injector: {repr(self.injector)}>'

    def __contains__(self, k):
        return k in self.injector

    async def __call__(self, cls, *args, **kwargs):
        '''Coroutine to Construct an instance of cls using the providers in this injector.
        Instantiate providers as needed.  In general a sub-injector is not
        constructed.  However if any keyword arguments pased in specify a
        dependency, then construct an injector for that.  Keyword arguments
        and arguments are passed to the class to construct the object.  If
        keyword arguments do specify a dependency, they must satisfy the
        InjectionKey involved.
'''
        if not hasattr(self, 'loop'):
            self.loop = self.get_instance(asyncio.AbstractEventLoop)
        try:
            res = self._instantiate(
                cls, *args, **kwargs,
                _loop=self.loop,
                _placement=None,
                _interim_placement=None,
            )

            if isinstance(res, (asyncio.Future, collections.abc.Coroutine)):
                return await res
            else:
                return res
        except Exception as e:
            if filter_tracebacks:
                tb_utils.filter_chatty_modules(e, _chatty_modules, 1)
            raise

    async def get_instance_async(self, k):
        futures = []
        try:
            res = self.get_instance(k,
                                    loop=self.loop,
                                    futures=futures)
            if isinstance(res, (asyncio.Future, collections.abc.Coroutine)):
                return await res
            else:
                return res
        except Exception as e:
            if filter_tracebacks:
                tb_utils.filter_chatty_modules(e, _chatty_modules, 1)
            raise

    async def filter_instantiate_async(self, target, predicate,
                                       *, stop_at=None,
                                       ready=None):

        '''
        Like :meth:`filter_instantiate` except in an async context.  With *filter_instance* objects are not-ready by default.  With *filter_instance_async*, the ready state is default (true unless in the middle of instantiating something with ready false).

        '''
        results = []
        result_keys = []
        for k in self.filter(target, predicate, stop_at=stop_at):
            if ready is not None:
                k = InjectionKey(k, _ready=ready)
            results.append(self.get_instance_async(k))
            result_keys.append(k)
        results = await asyncio.gather(*results)
        zipped = zip(result_keys, results)
        return [z for z in zipped if z[1] is not None]


async def _handle_async_deps(obj, cycle_set):
    if cycle_set is None:
        cycle_set = set()
    futures = []
    try:
        repr = repr(obj)
    except Exception:
        repr = f"<Error in repr for {obj.__class__.__name__}"
    for attr, dep in obj.__class__._injection_dependencies.items():
        if dep is None:
            continue
        if dep.ready is False:
            continue
        val = getattr(obj, attr, None)
        if val is None:
            continue
        if not isinstance(val, AsyncInjectable):
            continue
        if id(val) in cycle_set:
            import warnings
            warnings.warn(f'{obj}.async_become_ready: @inject({attr}={dep}) produced dependency cycle with {val}')
            continue
        cycle_set.add(id(val))
        future = asyncio.ensure_future(val.async_become_ready(cycle_set=cycle_set))
        future.set_name(f'Dependency become ready for {dep} of {repr}')
        futures.append(future)

    gather = asyncio.gather(*futures)
    await gather
    return await obj.async_ready()


# Injector cross reference support
class InjectorXrefMarkerMeta(type):

    def __repr__(self):
        return f'injector_xref({self.injectable_key}, {self.target_key})'


class InjectorXrefMarker(AsyncInjectable, metaclass=InjectorXrefMarkerMeta):

    # A separate subclass is created for each injector_xref with one
    # class attributes: target_key.  If possible our __new__ entirely
    # resolves to target_key and simply returns that object.  If that
    # requires asynchronous action, then we actually instantiate
    # ourself and depend on async_resolv.  We receive an injectable
    # through dependency injection.

    def __new__(cls, injectable):
        if isinstance(injectable, Injector):
            injector = injectable
        else:
            injector = injectable.injector
        try:
            res = injector.get_instance(cls.target_key)
            return res
        except AsyncRequired:
            # Don't pass in the injector; we don't want to claim it
            self = super().__new__(cls)
            self.ainjector = injector.get_instance(AsyncInjector)
            return self

    @classmethod
    def supplementary_injection_keys(cls, k):
        return
        yield None  # to make it a generator

    @classmethod
    def satisfies_injection_key(cls, k):
        target = cls.target_key.target
        if (isinstance(target, type) and issubclass(target, Injectable)) or \
           isinstance(target, Injectable):
            return target.satisfies_injection_key(k)
        return True

    async def async_resolve(self):
        return await self.ainjector.get_instance_async(self.target_key)


def injector_xref(injectable_key: InjectionKey,
                  target_key: InjectionKey,
                  ):
    '''

    Request that one injector look up a target in another injector.  Typical usage::

        base_injector.add_provider(target_key,
            injector_xref(sub_injector_key, target_key))

    :param injectable_key: The :class:`InjectionKey` of an :class:`Injectable` or an :class:`Injector` in which the target is actually looked up.  If *None*, then look up in the current injector.

    :param target_key: An :class:`InjectionKey` registered with the
    injector belonging to *injectable_key*.  It is important that
    *target_key* actually be provided by that injector.  In the common
    case where the parent of *injectable_key* eventually chains back
    to the injector in which this *injector_xref* is provided, a loop
    can happen otherwise.

    '''
    # If we return an AsyncInjectable that needs to be ready, our
    # caller will handle calling async_become_ready.  We want to
    # requesta not_ready object to avoid cycles.  If asynchronous is
    # called, we'll hold with an interim future providing the
    # dependency while InjectorXrefMarker's async_resolve runs.
    # That's necessary.  If we don't explicitly request _ready=False,
    # we will also hold with an interim future while the inner
    # async_become_ready is called.  In cases like AnsibleInventory
    # where someone asks for _ready=False while async_become_ready is
    # running, this creates a cycle.
    tkey = InjectionKey(target_key, _ready=False)
    ikey = injectable_key if injectable_key is not None else InjectionKey(Injector)

    @inject(injectable=ikey,
            injector=None)
    class instance(InjectorXrefMarker):
        target_key = tkey
        injectable_key = ikey
    return instance


async def shutdown_injector(injector, timeout=5):

    '''
    Close an injector and cancel running tasks


    This closes an injector, canceling any running tasks.  It waits up to *timeout* seconds for any canceled tasks to terminate.

'''
    canceled_futures = []
    injector.close(canceled_futures=canceled_futures)
    if canceled_futures:
        await asyncio.wait(canceled_futures, timeout=timeout)


def _call_close(obj, canceled_futures):
    if not hasattr(obj, 'close'):
        return
    sig = inspect.signature(obj.close)
    try:
        if 'canceled_futures' in sig.parameters:
            return obj.close(canceled_futures=canceled_futures)
        else:
            return obj.close()
    except TypeError:
        pass  # calling on not yet constructed class


def aspect_for(cls: typing.Type[Injectable],
               property: str):
    '''A decorator for a class indicating that the class provides an
    optional dependency to another class.  Once decorated, newly
    created instances of *cls* may have a property *property* if
    *cls*'s :class:`Injector` is able to instantiate the decorated
    class.  Usage might be something like::

        @aspect_for(Machine, 'model')
        class MachineModel:
        . . .

    Then in code that is given a machine you can do things like::

        try:
            os = machine.model.os
        except AttributeError: os = "unknown"

    In a dependency injection framework, it is common for a middle
    layer not to care about domain-specific knowledge that is needed
    both in an interface layer and in domain-specific implementation
    code.  As an example, implementations of
    :class:`~carthage.container.Container` don't care much what
    operating system the container is running.  However user-interface
    code to configure the container might need to set what operating
    system is desired, and various domain-specific
    :class:`~carthage.machine.BaseCustomization` classes may need to
    change behavior based on the operating system.  Aspects provide a
    convenient syntax for the injector hierarchy to be used to access
    this information.

    '''
    def wrapper(val):
        raise NotImplementedError("This approach is not great because of subclasses")
        if property in cls._injection_dependencies:
            raise TypeError(f'{k} is already a dependency for {cls.__name__}.')
        inject(property=val)(cls)
        return val
    return wrapper


def is_obj_ready(obj):
    if isinstance(obj, AsyncInjectable):
        return obj._async_ready_state == ReadyState.READY
    return True


@contextlib.contextmanager
def injection_failed_unlogged():
    '''Typically InjectionFailed is logged at severity error.  Within this context, InjectionFailed errors are not logged.
    '''
    reset = log_injection_failed.set(False)
    try:
        yield
    finally:
        log_injection_failed.reset(reset)

@contextlib.contextmanager
def instantiation_not_ready(ready=False):
    '''
    In this context, calling an :class:`Injector`, or calling methods like :meth:`Injector.get_instance` will not bring objects to ``async_ready`` state unless the *ready* parameter is True.

    This can be overridden in the key of a dependency.  For example::

        @inject_autokwargs(bar=InjectionKey(Bar, _ready=True))
        class Foo(AsyncInjectable):

            async def async_ready(self): pass # do stuff in this method

    And then later::

        with instantiation_not_ready():
            foo = await ainjector(Foo)

    The *bar* dependency of *foo* will be brought to *async_ready* state.  However, *foo* itself will not be ready: *foo.async_ready()* will not be called.

    Do not confuse this function with :var:`instantitate_to_ready`.  This is a context manager to manipulate the desired ready state.  *instantiate_to_ready* is an internal context variable that tracks desired ready state.
    Calling ``instantiation_not_ready(ready=True)`` may be a bit counter-intuitive; it will reverse the effect of calling ``instantiation_not_ready()`` and turn on default ready state.
    
    '''
    reset = instantiate_to_ready.set(ready)
    try:
        yield
    finally:
        instantiate_to_ready.reset(reset)


async def resolve_deferred(ainjector, item, args:dict):
    '''
    Often it is desirable to have a data structure that has some elements dynamically calculated.  For example looking up something in the environment of an injector, or calling some function that takes dependencies from an injector.
    Examples of where this is used include :class:`~carthage.network.NetworkConfig`.

    This function traverses a data structure starting at an atomic element, list, or dict.  It returns a copy where:

    *  :class:`InjectionKey` are replaced with :meth:`AsyncInjector.get_instance_async` in the context of the current injector

    * A *callable* is called in the context of the current injector.  If the callable has a signature, and if items in *args* match keys in the signature, the value from *args* is plugged into the call.  An injected dependency could be used instead, but in some cases the performance of setting up an injector just for one parameter value doesn't make sense.
    

    * dicts, lists, and tuples are recursed, although tuples become lists.

    * Other items are not modified.

    This function does not deal with recursive data structures.
    '''
    # This effectively uses cps internally.  The result argument to the inner functions is a callback that takes the result and puts it in the right place.
    def handle_list(l, result):
        def place_list(idx):
            def place(elt):
                out[idx] = elt
            return place
        out = [None]*len(l)
        for i, elt in enumerate(l):
            handle(elt, place_list(i))
        result(out)
    def handle_dict(d, result):
        def place_dict(k):
            def place(elt):
                out[k] = elt
            return place
        out = {k:None for k in d} # preserve order
        for k,v in d.items():
            handle(v, place_dict(k))
        result(out)
    def handle_injection(k, result):
        fut = asyncio.ensure_future(ainjector.get_instance_async(k))
        def done(f):
            result(f.result())
        fut.add_done_callback(done)
        futures.append(fut)
    def handle_callable(c, result):
        try:
            signature = inspect.signature(c)
            keys = signature.parameters.keys()
        except Exception: keys = set()
        in_args = {}
        for k in set(keys)& args_keys:
            in_args[k] = args[k]
        fut = asyncio.ensure_future(ainjector(c, **in_args))
        def done(f):
            result(f.result())
        fut.add_done_callback(done)
        futures.append(fut)
    def handle(elt, result):
        if isinstance(elt, dict):
            handle_dict(elt, result)
        elif isinstance(elt, (list,tuple)):
            handle_list(elt, result)
        elif isinstance(elt, InjectionKey):
            handle_injection(elt, result)
        elif callable(elt):
            handle_callable(elt, result)
        else:result(elt)
    futures = []
    args_keys = frozenset(args.keys())
    def base_result(elt):
        nonlocal to_return
        to_return = elt
    to_return = None
    handle(item, base_result)
    if futures: await asyncio.gather(*futures)
    return to_return


__all__ = [
    'AsyncInjectable', 'AsyncInjector', 'AsyncRequired',
    'DependencyProvider',
    'ExistingProvider', 'Injectable', 'InjectionFailed',
    'InjectionKey', 'Injector', 'InstantiationContext', 'aspect_for',
    'NotPresent',
    'dependency_quote', 'inject',
    'inject_autokwargs', 'injector_xref',
    'partial_with_dependencies', 'shutdown_injector',
    'resolve_deferred',
    'injection_failed_unlogged', 'instantiation_not_ready']



# Copyright (C) 2019, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import pytest
import os.path
import yaml
from carthage.config import *
from carthage.dependency_injection import Injector, inject
resource_dir = os.path.dirname(__file__)


def test_add_config(ainjector):
    injector = ainjector.injector

    class Defaults(ConfigSchema, prefix=""):
        foo: int = 33
    cl = injector(ConfigLayout)
    assert cl.foo == 33


def test_override_config(ainjector):
    injector = ainjector.injector
    injector.replace_provider(ConfigLayout)

    @inject(cl=ConfigLayout)
    def fn(cl):
        assert cl.hadron_operations == 99
    injector.add_provider(config_key("hadron_operations"), 99)
    injector(fn)


def test_substitution_in_yaml(ainjector):
    injector = ainjector.injector(Injector)
    cl = injector(ConfigLayout)
    cl.load_yaml(open(os.path.join(resource_dir, "override-config.yml"), 'rt'))
    assert cl.hadron_operations == f"{cl.base_dir}/hadron-operations"


def test_list_in_yaml(ainjector):
    class ConfigList(ConfigSchema, prefix=""):
        l: list = (2, 3)
    injector = ainjector.injector(Injector)
    cl = injector(ConfigLayout)
    assert tuple(cl.l) == (2, 3)
    cl.load_yaml(yaml.dump(dict(
        l=[9, 10],
    )), path=".")
    assert tuple(cl.l) == (9, 10)



# Copyright (C) 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import os
from glob import glob
from carthage.setup_tasks import setup_task, SetupTaskMixin, cross_object_dependency, SkipSetupTask
from carthage.dependency_injection import *
from carthage import ConfigLayout
from carthage.machine import FilesystemCustomization
from carthage.utils import memoproperty

from carthage import sh


@inject(config=ConfigLayout)
class ExtraRepository(SetupTaskMixin, AsyncInjectable):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.stamp_path = self.config.extra_packages.repository_dir

    @memoproperty
    def path(self):
        return self.config.extra_packages.repository_dir

    @setup_task("Create Repository")
    def create_repository(self):
        repo_dir = self.config.extra_packages.repository_dir
        conf_dir = os.path.join(repo_dir, 'conf')
        config = self.config
        os.makedirs(conf_dir, exist_ok=True)
        with open(os.path.join(conf_dir, 'distributions'), 'wt') as f:
            f.write(f'''\
Codename: carthage_extra
Suite: carthage_extra
Architectures: {config.extra_packages.architectures}
Components: {config.extra_packages.components}

'''
                    )

    @setup_task("Add Packages to Repository")
    async def add_packages(self):
        packages_dir = self.config.extra_packages.packages_dir
        repo_dir = self.config.extra_packages.repository_dir
        for c in glob(os.path.join(packages_dir, '*.changes')):
            await sh.reprepro(
                "--ignore=wrongdistribution",
                "include", "carthage_extra",
                c,
                _bg=True, _bg_exc=True,
                _cwd=repo_dir)

    @add_packages.invalidator()
    def add_packages(self, **kwargs):
        last_run = self.check_stamp(self.__class__.add_packages.stamp)[0]
        if not last_run:
            return False
        packages_dir = self.config.extra_packages.packages_dir
        for c in glob(os.path.join(packages_dir, "*.changes")):
            st = os.stat(c)
            if st.st_mtime > last_run:
                return False
        return True  # Not invalidated

    def source_lines(self):
        repository_dir = self.config.extra_packages.repository_dir
        if os.path.exists(os.path.join(
                repository_dir, "dists")):
            return(f'''
deb [trusted=yes] file:///extra_packages carthage_extra {self.config.extra_packages.components}
''')
        return ""


@inject(repo=ExtraRepository)
class AddExtraRepo(FilesystemCustomization):

    description = "Add extra repository to machine"

    repo_dependency = cross_object_dependency(ExtraRepository.add_packages, "repo")

    @setup_task("Create Extra Repository Sources")
    def manage_sources(self):
        if not self.repo.source_lines():
            try:
                os.unlink(os.path.join(self.path,
                                       "etc/apt/sources.list.d/extra_repo.list"))
            except FileNotFoundError:
                pass
            raise SkipSetupTask()
        with open(os.path.join(self.path, "etc/apt/sources.list.d/extra_repo.list"), "wt") as f:
            f.write(self.repo.source_lines())
        with open(os.path.join(self.path, "etc/apt/preferences.d/10-extra.pref"), "wt") as f:
            f.write('''\
package: *
pin: release n=carthage_extra
pin-priority: 1030
''')

    @setup_task("Copy in extra repo")
    async def copy_in_repo(self):
        j = os.path.join
        if not self.repo.source_lines():
            raise SkipSetupTask()
        await sh.rsync(
            '-a',
            '--delete-delay',
            j(self.repo.path, "dists"),
            j(self.repo.path, "pool"),
            j(self.path, "extra_packages"),
            _bg=True,
            _bg_exc=False)


@inject(repo=ExtraRepository)
class UpgradeFromRepo(FilesystemCustomization):

    description = "Schedule apt update/apt full-upgrade"

    repo_dependency = cross_object_dependency(ExtraRepository.add_packages, 'repo')

    @setup_task('update packages')
    async def update_packages(self):
        await self.run_command('apt', 'update',
                               _bg=True, _bg_exc=False)

    @setup_task("upgrade packages")
    async def upgrade_packages(self):
        await self.run_command(
            'sh', '-c',
            'DEBIAN_FRONTEND=NONINTERACTIVE apt --allow-downgrades -y full-upgrade',
            _bg=True,
            _bg_exc=False)



# Copyright (C) 2018, 2019, 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import functools
import inspect
import json
import pytest
import sys
import time

from .dependency_injection import inject, InjectionKey

from _pytest.reports import TestReport
from _pytest.nodes import Node


'''
Decorators and functions for use with Carthage.

This module is typically used alongside the :ref:`carthage.pytest_plugin` Pytest plugin.  This module contains decorators like :funcref:`async_test` and that plugin provides fixtures and hooks needed for these items to work.  Supported functionality includes:

* *asyncio* tests that support :ref:`carthage.inject` style dependency injection

* Running a subtest within an item implementing :ref:`SshMixin` and collecting the results.

'''


class TestTiming:

    time_remaining = 120
    previous_timing = None
    __test__ = False

    '''
    Usage::

         with TestTiming(300):
              # Block of code that is given 300 seconds to run

'''

    def __init__(self, timing: float):
        self.time_remaining = timing
        self.initial_time_remaining = timing

    def __enter__(self):
        global test_timing
        self.previous_timing = test_timing
        test_timing = self

    def __exit__(self, *args):
        global test_timing
        test_timing = self.previous_timing


test_timing = None


def time_loop(loop, future):
    '''Run *future* in 5 second slices, adjusting time_remaining.  If time runs out, cancel the future but eventually raise TimeoutError.
'''
    done = tuple()
    #: None or the eventual error message to print.
    timed_out = None
    time_delta = 5
    while not done:
        now = time.time()
        time_remaining = test_timing.time_remaining
        done, pending = loop.run_until_complete(asyncio.wait(
            [future], timeout=time_remaining if time_remaining < time_delta else time_delta))
        # test_timing.time_remaining may have increased while running.
        if not done:
            test_timing.time_remaining -= time.time() - now
            if test_timing.time_remaining < 0:
                if timed_out:
                    raise TimeoutError(timed_out)
                timed_out = f'{future} failed to complete in {test_timing.initial_time_remaining} seconds'
                future.cancel(msg='timed out')
                test_timing.time_remaining = 30
    if timed_out:
        if not future.cancelled():
            raise ValueError('Task failed to cancel on timeout')
        raise TimeoutError(timed_out)


def async_test(t):
    '''A decorator for wrapping a test. *t* is expected to be a coroutine
    and will be run inside an event loop.  The test may take either
    Carthage style dependencies using the :funcref:`carthage.inject`
    decorator or Pytest style fixtures as function arguments.  Any
    Carthage style injected item will be removed from the list of
    Pytest style injectors.

    '''
    # This is messy because Pytest's introspection logic does not
    # respect inspect.Signature.  So we don't actually include the
    # _wrapped_ item and explicitly mark tests with pytest.usefixture
    # to indicate which fixtures are required.  However, we need to
    # depend on a pytest_collect_modifyitems hook in
    # carthage.pytest_plugin to make the fixtures available to the
    # function.

    sig = inspect.signature(t)
    orig_loop = True
    orig_ainjector = True

    @functools.wraps(t)
    def wrapper(loop, *args, **kwargs):
        global test_timing
        test_timing = TestTiming(TestTiming.time_remaining)

        if orig_loop:
            kwargs['loop'] = loop
        ainjector = kwargs.get('ainjector', None)
        if ainjector is None:
            task = asyncio.ensure_future(t(*args, **kwargs), loop=loop)
        else:
            if not orig_ainjector:
                del kwargs['ainjector']
            task = asyncio.ensure_future(ainjector(t, *args, **kwargs), loop=loop)
        time_loop(loop, task)
        return task.result()
    params = list(sig.parameters.values())
    try:
        params = list(filter(lambda p: p.name not in t._injection_dependencies, params))
    except AttributeError:
        pass  # no @inject call
    param_names = set(p.name for p in params)
    if 'loop' not in param_names:
        orig_loop = False
        params.append(inspect.Parameter(name="loop", kind=inspect._KEYWORD_ONLY))
    if 'ainjector' not in param_names:
        orig_ainjector = False
        if hasattr(t, '_injection_dependencies'):
            params.append(inspect.Parameter(name="ainjector", kind=inspect._KEYWORD_ONLY))
    sig = sig.replace(parameters=params)
    wrapper.__signature__ = sig
    pytest.mark.usefixtures(*sig.parameters.keys())(wrapper)
    del wrapper.__dict__['__wrapped__']
    wrapper.place_as = t
    return wrapper


_test_results_serial = 0


async def subtest_controller(request, target, pytest_args,
                             python_path="", ssh_agent=False):
    '''Ssh into a given machine using a :class:`carthage.SshMixin` and
    run a series of pytests.  This is typically run by a :ref:`test
    controller` from within a test on the test controller.  This
    function arranges for the tests to be run and collects the
    results.  The results are reported as inferior to the test
    context represented by *request*.

    :param: request
        A ``pytest`` request fixture representing the test that is the **test controller**

    :param: target
        A :class:`carthage.SshMixin` with ``pytest`` installed and available and the ``carthage.pytest_plugin`` available.

    :param: pytest_args
        A list of arguments to passed into pytest on the target system.

    :param python_path:
        Set the remote *PYTHONPATH* environment variable to this value.  Typically used when a set of tests is copied to the system to point to directory containings tests.

    :param ssh_agent:
        If *True*, then forward agent credentials.

'''
    if isinstance(pytest_args, str):
        pytest_args = [pytest_args]
    json_frag = f'/tmp/{id(pytest_args)}.json'
    pytest_args = ['--carthage-json=' + json_frag] + pytest_args
    ssh_args = []
    if ssh_agent:
        ssh_args.append('-A')
    if python_path:
        ssh_args.append('PYTHONPATH=' + python_path)
    await target.ssh(*ssh_args,
                     'pytest-3', *pytest_args,
                     _bg=True, _bg_exc=False,
                     _out=sys.stdout)
    json_out = await target.ssh('cat', json_frag)
    report_list = json.loads(json_out.stdout)
    for i in report_list:
        try:
            n = Node.from_parent(name=i['nodeid'], parent=request.node)
        except AttributeError:
            n = Node(name=i['nodeid'], parent=request.node)
            i['nodeid'] = n.nodeid
        report = TestReport(**i)
        capmanager = request.config.pluginmanager.getplugin("capturemanager")
        with capmanager.global_and_fixture_disabled():
            n.ihook.pytest_runtest_logreport(report=report)


__all__ = ['async_test', 'subtest_controller', 'TestTiming']
